from datetime import timedelta, timezone
from logging import getLogger
from queue import Queue
from typing import Callable, Collection, Dict, Final, List, Optional, Type
from uuid import UUID

from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData

from platypush.context import get_bus
from platypush.entities.bluetooth import BluetoothDevice, BluetoothService
from platypush.message.event.bluetooth import (
    BluetoothDeviceConnectedEvent,
    BluetoothDeviceDisconnectedEvent,
    BluetoothDeviceFoundEvent,
    BluetoothDeviceSignalUpdateEvent,
    BluetoothDeviceEvent,
)
from platypush.utils import utcnow

from .._cache import EntityCache
from .._model import ServiceClass
from .._plugins import BaseBluetoothPlugin
from .._types import DevicesBlacklist
from ._cache import DeviceCache
from ._mappers import device_to_entity

_rssi_update_interval: Final[float] = 30.0
""" How often to trigger RSSI update events for a device. """

_excluded_manufacturers: Final[Collection[str]] = {
    'Apple Continuity',
    'Apple, Inc.',
    'GAEN',
    'Google',
    'Google Inc.',
    'Microsoft',
    'Samsung Electronics Co., Ltd',
}
"""
Exclude beacons from these device manufacturers by default (main offenders
when it comes to Bluetooth device space pollution).
"""

logger = getLogger(__name__)


def _has_changed(
    old: Optional[BluetoothDevice],
    new: BluetoothDevice,
    attr: str,
    tolerance: Optional[float] = None,
) -> bool:
    """
    Returns True if the given attribute has changed on the device.
    """
    if old is None:
        return False  # No previous value

    old_value = getattr(old, attr)
    new_value = getattr(new, attr)
    if tolerance and (old_value is not None and new_value is not None):
        return abs(new_value - old_value) < tolerance
    return old_value != new_value


def _has_been_set(
    old: Optional[BluetoothDevice], new: BluetoothDevice, attr: str, value: bool
) -> bool:
    """
    Returns True if the given attribute has changed and its new value matches
    the given value.
    """
    if not _has_changed(old, new, attr):
        return False

    new_value = getattr(new, attr)
    return new_value == value


event_matchers: Dict[
    Type[BluetoothDeviceEvent],
    Callable[[Optional[BluetoothDevice], BluetoothDevice], bool],
] = {
    BluetoothDeviceConnectedEvent: lambda old, new: _has_been_set(
        old, new, 'connected', True
    ),
    BluetoothDeviceDisconnectedEvent: lambda old, new: old is not None
    and bool(old.connected)
    and _has_been_set(old, new, 'connected', False),
    BluetoothDeviceFoundEvent: lambda old, new: old is None
    or (old.reachable is False and new.reachable is True),
    BluetoothDeviceSignalUpdateEvent: lambda old, new: (  # type: ignore
        (new.rssi is not None or new.tx_power is not None)
        and bool(old and old.rssi)
        and (
            _has_changed(old, new, 'rssi', tolerance=5)
            or _has_changed(old, new, 'tx_power', tolerance=5)
        )
        and (
            not (old and old.updated_at)
            or utcnow() - old.updated_at.replace(tzinfo=timezone.utc)
            >= timedelta(seconds=_rssi_update_interval)
        )
    ),
}
""" A static ``BluetoothDeviceEvent -> MatchCallback`` mapping. """


# pylint: disable=too-few-public-methods
class EventHandler:
    """
    Event handler for BLE devices.
    """

    def __init__(
        self,
        device_queue: Queue,
        device_cache: DeviceCache,
        entity_cache: EntityCache,
        plugins: Collection[BaseBluetoothPlugin],
        exclude_known_noisy_beacons: bool,
        blacklist: DevicesBlacklist,
    ):
        """
        :param device_queue: Queue used to publish updated devices upstream.
        :param device_cache: Device cache.
        :param entity_cache: Entity cache.
        :param exclude_known_noisy_beacons: Exclude known noisy beacons.
        :param blacklist: Blacklist rules.
        """
        self._device_queue = device_queue
        self._device_cache = device_cache
        self._entity_cache = entity_cache
        self._plugins = plugins
        self._exclude_known_noisy_beacons = exclude_known_noisy_beacons
        self._blacklist = blacklist

    def __call__(self, device: BLEDevice, data: AdvertisementData):
        """
        Handler for Bluetooth device advertisement packets.

        1. It generates the relevant
            :class:`platypush.message.event.bluetooth.BluetoothDeviceEvent` if the
            state of the device has changed.

        2. It builds the relevant
            :class:`platypush.entity.bluetooth.BluetoothDevice` entity object
            populated with children entities that contain the supported
            properties.

        3. Publishes the updated entity to the upstream queue.

        :param device: The Bluetooth device.
        :param data: The advertised data.
        """

        new_entity = device_to_entity(device, data)
        if self._exclude_known_noisy_beacons and self._is_noisy_beacon(new_entity):
            logger.debug(
                'exclude_known_noisy_beacons is set to True: skipping beacon from device %s',
                device.address,
            )
            return

        if self._blacklist.matches(new_entity):
            logger.debug('Ignoring blacklisted device: %s', device.address)
            return

        # Extend the new entity with children entities added by the plugins
        for plugin in self._plugins:
            plugin.extend_device(new_entity)

        events: List[BluetoothDeviceEvent] = []
        existing_entity = self._entity_cache.get(device.address)
        events += [
            event_type.from_device(new_entity)
            for event_type, matcher in event_matchers.items()
            if matcher(existing_entity, new_entity)
        ]

        self._device_cache.add(device)
        for event in events:
            get_bus().post(event)

        if events:
            new_entity.reachable = True  # type: ignore
            self._device_queue.put_nowait(new_entity)

    @staticmethod
    def _is_noisy_beacon(device: BluetoothDevice) -> bool:
        """
        Check if the beacon received from the given device should be skipped.
        """
        # Exclude iBeacons
        if device.model == 'iBeacon':
            return True

        # "Noisy" beacon devices usually have no associated friendly name. If a
        # device has a valid name, we should probably include it.
        if (
            device.name
            and device.name.replace('-', ':').lower() != device.address.lower()
        ):
            return False

        # If the manufacturer is in the excluded list, we should skip it
        if (
            device.manufacturer in _excluded_manufacturers
            or device.model in _excluded_manufacturers
        ):
            return True

        # If the device has no children and no manufacturer, skip it
        if not (device.children and device.manufacturer):
            return True

        # If the device has any children other than services, don't skip it
        if any(not isinstance(child, BluetoothService) for child in device.children):
            return False

        # If the device's only children are unknown services, skip it
        if not any(
            isinstance(child, BluetoothService)
            and child.service_class != ServiceClass.UNKNOWN
            for child in device.children
        ):
            return True

        mapped_uuids = [
            (
                int(str(srv.uuid).split('-', maxsplit=1)[0], 16) & 0xFFFF
                if isinstance(srv.uuid, UUID)
                else srv.uuid
            )
            for srv in device.services
        ]

        # If any of the services matches the blacklisted manufacturers, skip
        # the event.
        return any(
            str(ServiceClass.get(uuid)) in _excluded_manufacturers
            for uuid in mapped_uuids
        )



import datetime as dt
import os
import pathlib
import re
import stat
import subprocess
import time
from threading import RLock
from typing import List, Optional
from urllib.parse import quote_plus

import requests
from sqlalchemy import create_engine, text

from platypush.config import Config
from platypush.context import Variable

from .._model import TorrentSearchResult
from ._base import TorrentsCsvBaseProvider
from ._constants import TORRENTS_CSV_URL_LAST_CHECKED_VAR

SQL_INIT_TEMPLATE = """
create table torrent_tmp (
  infohash text primary key,
  name text not null,
  size_bytes integer not null,
  created_unix integer(4) not null,
  seeders integer not null,
  leechers integer not null,
  completed integer not null,
  scraped_date integer(4) not null,
  published integer(4) not null
);

.separator ,
.import --skip 1 '{csv_file}' torrent_tmp

create index idx_name on torrent_tmp(lower(name));
create index idx_seeders on torrent_tmp(seeders);
create index idx_created_unix on torrent_tmp(created_unix);

drop table if exists torrent;
alter table torrent_tmp rename to torrent;
"""


class TorrentsCsvLocalProvider(TorrentsCsvBaseProvider):
    """
    This class is responsible for managing a local checkout of the torrents-csv
    dataset.
    """

    def __init__(
        self,
        download_csv: bool,
        csv_url: str,
        csv_url_check_interval: int,
        csv_path: Optional[str] = None,
        db_path: Optional[str] = None,
        **kwargs,
    ):
        """
        Note that at least one among ``download_csv``, ``csv_path`` and ``db_path``
        should be provided.

        :param download_csv: If True then the CSV file will be downloaded from the
            specified ``csv_url``.
        :param csv_url: The URL from which the CSV file will be downloaded.
        :param csv_url_check_interval: The interval in seconds after which the CSV
            should be checked for updates.
        :param csv_path: The path to the CSV file. If not provided, and download_csv
            is set to True, then the CSV file will be downloaded to
            ``<WORKDIR>/torrent/torrents.csv``.
        :param db_path: The path to the SQLite database. If not provided, and
            ``csv_path`` or ``download_csv`` are set, then the database will be created
            from a local copy of the CSV file.
        """
        super().__init__(**kwargs)
        assert (
            download_csv or csv_path or db_path
        ), 'You must provide either download_csv, csv_path or db_path'

        self._init_csv_lock = RLock()
        self._init_db_lock = RLock()
        self._csv_url_check_interval = csv_url_check_interval

        if download_csv:
            csv_path = (
                os.path.expanduser(csv_path)
                if csv_path
                else os.path.join(Config.get_workdir(), 'torrent', 'torrents.csv')
            )

            with self._init_csv_lock:
                self._download_csv(csv_url=csv_url, csv_path=csv_path)

        if csv_path:
            db_path = (
                os.path.expanduser(db_path)
                if db_path
                else os.path.join(os.path.dirname(csv_path), 'torrents.db')
            )

            with self._init_db_lock:
                self._build_db(csv_path=csv_path, db_path=db_path)

        assert db_path, 'No download_csv, csv_path or db_path provided'
        assert os.path.isfile(db_path), f'Invalid db_path: {db_path}'
        self.db_path = db_path

    def _get_engine(self):
        return create_engine(
            'sqlite:///' + ('/'.join(map(quote_plus, self.db_path.split(os.path.sep))))
        )

    def _download_csv(self, csv_url: str, csv_path: str):
        if not self._should_download_csv(
            csv_url=csv_url,
            csv_path=csv_path,
            csv_url_check_interval=self._csv_url_check_interval,
        ):
            return

        self.logger.info(
            'Downloading the torrents CSV file from %s to %s', csv_url, csv_path
        )

        response = requests.get(csv_url, stream=True, timeout=60)
        response.raise_for_status()
        size = int(response.headers.get('Content-Length', 0))
        torrents_csv_dir = os.path.dirname(csv_path)
        pathlib.Path(torrents_csv_dir).mkdir(parents=True, exist_ok=True)

        with open(csv_path, 'wb') as f:
            written = 0

            for chunk in response.iter_content(chunk_size=8192):
                f.write(chunk)
                written += len(chunk)
                if size:
                    percent = 100.0 * written / size
                    prev_percent = max(0, 100.0 * (written - len(chunk)) / size)
                    if round(percent / 5) > round(prev_percent / 5):
                        self.logger.info('... %.2f%%\r', percent)

        self.logger.info('Downloaded the torrents CSV file to %s', csv_path)

    def _build_db(self, csv_path: str, db_path: str):
        if not self._should_update_db(csv_path, db_path):
            return

        self.logger.info(
            'Refreshing SQLite database %s from CSV file %s', db_path, csv_path
        )

        db_dir = os.path.dirname(db_path)
        pathlib.Path(db_dir).mkdir(parents=True, exist_ok=True)

        with subprocess.Popen(
            ['sqlite3', db_path], stdin=subprocess.PIPE, text=True
        ) as proc:
            proc.communicate(SQL_INIT_TEMPLATE.format(csv_file=csv_path))

        self.logger.info(
            'Refreshed SQLite database %s from CSV file %s: ready to search',
            db_path,
            csv_path,
        )

    @staticmethod
    def _should_update_db(csv_path: str, db_path: str) -> bool:
        if not os.path.isfile(csv_path):
            return False

        if not os.path.isfile(db_path):
            return True

        return os.stat(db_path)[stat.ST_MTIME] < os.stat(csv_path)[stat.ST_MTIME]

    def _should_download_csv(
        self, csv_url: str, csv_path: str, csv_url_check_interval: int
    ) -> bool:
        if not os.path.isfile(csv_path):
            self.logger.info('CSV file %s not found, downloading it', csv_path)
            return True

        if not self._should_check_csv_url(csv_url_check_interval):
            self.logger.debug('No need to check the CSV URL %s', csv_url)
            return False

        request = requests.head(csv_url, timeout=10)
        request.raise_for_status()
        last_modified_hdr = request.headers.get('Last-Modified')
        Variable(TORRENTS_CSV_URL_LAST_CHECKED_VAR).set(time.time())

        if not last_modified_hdr:
            self.logger.debug(
                "No Last-Modified header found in the CSV URL, can't compare thus downloading"
            )
            return True

        return (
            time.mktime(time.strptime(last_modified_hdr, '%a, %d %b %Y %H:%M:%S %Z'))
            > os.stat(csv_path)[stat.ST_MTIME]
        )

    @staticmethod
    def _should_check_csv_url(csv_url_check_interval: int) -> bool:
        last_checked = round(
            float(Variable(TORRENTS_CSV_URL_LAST_CHECKED_VAR).get() or 0)
        )
        return bool(
            csv_url_check_interval
            and time.time() - last_checked > csv_url_check_interval
        )

    def _search(
        self, query: str, *_, limit: int, page: int, **__
    ) -> List[TorrentSearchResult]:
        self.logger.debug(
            "Searching for %r on %s, limit=%d, page=%d",
            query,
            self.db_path,
            limit,
            page,
        )

        tokens = re.split(r'[^\w]', query.lower())
        where = ' and '.join(
            f'lower(name) like :token{i}' for i, _ in enumerate(tokens)
        )
        tokens = {f'token{i}': f'%{token}%' for i, token in enumerate(tokens)}

        with self._get_engine().connect() as conn:
            self.logger.debug('Connected to the database: %s', conn.engine.url)
            results = conn.execute(
                text(
                    f"""
                    select infohash, name, size_bytes, seeders, leechers, created_unix
                    from torrent
                    where {where}
                    order by seeders desc, created_unix desc
                    limit :limit
                    offset :offset
                    """
                ),
                {
                    **tokens,
                    'limit': max(int(limit), 0),
                    'offset': max(int(limit * (page - 1)), 0),
                },
            ).all()

        self.logger.debug('Found %d results', len(results))
        return [
            TorrentSearchResult(
                title=result[1],
                url=self._to_magnet(
                    info_hash=result[0],
                    torrent_name=result[1],
                ),
                size=result[2],
                seeds=int(result[3] or 0),
                peers=int(result[4] or 0),
                created_at=(
                    dt.datetime.fromtimestamp(result[5]).replace(tzinfo=dt.timezone.utc)
                    if result[5]
                    else None
                ),
            )
            for result in results
        ]



from dataclasses import dataclass
from enum import Enum
from typing import Union

DeviceType = Union[int, str]


@dataclass
class AudioDevice:
    """
    Maps the properties of an audio device.
    """

    index: int
    name: str
    hostapi: int
    max_input_channels: int
    max_output_channels: int
    default_samplerate: int
    default_low_input_latency: float = 0
    default_low_output_latency: float = 0
    default_high_input_latency: float = 0
    default_high_output_latency: float = 0


class AudioState(Enum):
    """
    Audio states.
    """

    STOPPED = 'STOPPED'
    RUNNING = 'RUNNING'
    PAUSED = 'PAUSED'


class StreamType(Enum):
    """
    Stream types.
    """

    INPUT = 'input'
    OUTPUT = 'output'



from logging import getLogger
from threading import Thread, Event, get_ident
from time import time
from typing import Dict, Optional

from platypush.context import get_bus
from platypush.entities import Entity
from platypush.message.event.entities import EntityUpdateEvent

from platypush.entities._base import EntityKey, EntitySavedCallback
from platypush.entities._engine.queue import EntitiesQueue
from platypush.entities._engine.repo import EntitiesRepository
from platypush.utils import get_remaining_timeout


class EntitiesEngine(Thread):
    """
    This thread runs the "brain" of the entities data persistence logic.

    Its purpose is to:

        1. Consume entities from a queue (synchronized with the upstream
           integrations that produce/handle them). The producer/consumer model
           ensure that only this thread writes to the database, packs events
           together (preventing excessive writes and throttling events), and
           prevents race conditions when SQLite is used.
        2. Merge any existing entities with their newer representations.
        3. Update the entities' taxonomy.
        4. Persist the new state to the entities' database.
        5. Trigger events for the updated entities.

    """

    def __init__(self) -> None:
        obj_name = self.__class__.__name__
        super().__init__(name=obj_name)

        self.logger = getLogger(name=obj_name)
        self._should_stop = Event()
        """ Event used to synchronize stop events downstream."""
        self._running = Event()
        """
        Event used to synchronize other threads to wait for the engine to
        start.
        """
        self._queue = EntitiesQueue(stop_event=self._should_stop)
        """ Queue where all entity upsert requests are received."""
        self._repo = EntitiesRepository()
        """ The repository of the processed entities. """
        self._callbacks: Dict[EntityKey, EntitySavedCallback] = {}
        """ (external_id, plugin) -> callback mapping"""

    def post(self, *entities: Entity, callback: Optional[EntitySavedCallback] = None):
        if callback:
            for entity in entities:
                self._callbacks[entity.entity_key] = callback

        self._queue.put(*entities)

    def wait_start(self, timeout: Optional[float] = None) -> None:
        started = self._running.wait(timeout=timeout)
        if not started:
            raise TimeoutError(
                f'Timeout waiting for {self.__class__.__name__} to start.'
            )

    @property
    def should_stop(self) -> bool:
        return self._should_stop.is_set()

    def stop(self):
        self._should_stop.set()

    def wait_stop(self, timeout: Optional[float] = None):
        start = time()
        stopped = self._should_stop.wait(
            timeout=get_remaining_timeout(timeout=timeout, start=start)
        )

        if not stopped:
            raise TimeoutError(
                f'Timeout waiting for {self.__class__.__name__} to stop.'
            )

        if get_ident() != self.ident:
            self.join(timeout=get_remaining_timeout(timeout=timeout, start=start))

    def notify(self, *entities: Entity):
        """
        Trigger an EntityUpdateEvent if the entity has been persisted, or queue
        it to the list of entities whose notifications will be flushed when the
        session is committed. It will also invoke any registered callbacks.
        """
        for entity in entities:
            get_bus().post(EntityUpdateEvent(entity=entity))
            self._process_callback(entity)

    def _process_callback(self, entity: Entity) -> None:
        """
        Process the callback for the given entity.
        """
        callback = self._callbacks.pop(entity.entity_key, None)
        if callback:
            try:
                callback(entity)
            except Exception as e:
                self.logger.error(
                    'Error while notifying updates for entity ID %d via %s: %s',
                    entity.id,
                    callback,
                    e,
                )
                self.logger.exception(e)

    def run(self):
        super().run()
        self.logger.info('Started entities engine')
        self._running.set()

        try:
            while not self.should_stop:
                # Get a batch of entity updates forwarded by other integrations
                entities = self._queue.get()
                if not entities or self.should_stop:
                    continue

                # Store the batch of entities
                try:
                    entities = self._repo.save(*entities)
                except Exception as e:
                    self.logger.error('Error while processing entity updates: %s', e)
                    self.logger.exception(e)
                    continue

                # Trigger EntityUpdateEvent events
                self.notify(*entities)
        finally:
            self.logger.info('Stopped entities engine')
            self._running.clear()



from platypush.message.event import Event


class HidBaseEvent(Event):
    """
    Base class for HID events.
    """

    def __init__(
        self,
        *args,
        path: str,
        serial_number: str,
        vendor_id: int,
        product_id: int,
        product_string: str,
        manufacturer_string: str,
        **kwargs
    ):
        super().__init__(
            *args,
            path=path,
            serial_number=serial_number,
            vendor_id=vendor_id,
            product_id=product_id,
            product_string=product_string,
            manufacturer_string=manufacturer_string,
            **kwargs
        )


class HidDeviceConnectedEvent(HidBaseEvent):
    """
    Event triggered when a device is discovered.
    """


class HidDeviceDisconnectedEvent(HidBaseEvent):
    """
    Event triggered when a device is disconnected.
    """


class HidDeviceDataEvent(HidBaseEvent):
    """
    Event triggered when a monitored device sends some data.
    """

    def __init__(self, *args, data: str, **kwargs):
        """
        :param data: Hex-encoded representation of the received data.
        """
        super().__init__(*args, data=data, **kwargs)


# vim:sw=4:ts=4:et:



from marshmallow import EXCLUDE, fields, post_dump, post_load, pre_dump, pre_load
from marshmallow.schema import Schema

from platypush.plugins.media import PlayerState
from platypush.schemas import DateTime


class MopidyTrackSchema(Schema):
    """
    Mopidy track schema.
    """

    uri = fields.String(required=True, metadata={"description": "Track URI"})
    file = fields.String(
        metadata={"description": "Track URI, for MPD compatibility purposes"}
    )
    artist = fields.String(missing=None, metadata={"description": "Artist name"})
    title = fields.String(missing=None, metadata={"description": "Track title"})
    album = fields.String(missing=None, metadata={"description": "Album name"})
    artist_uri = fields.String(
        missing=None, metadata={"description": "Artist URI (if available)"}
    )
    album_uri = fields.String(
        missing=None, metadata={"description": "Album URI (if available)"}
    )
    time = fields.Float(
        missing=None, metadata={"description": "Track length (in seconds)"}
    )
    playlist_pos = fields.Integer(
        missing=None,
        metadata={"description": "Track position in the tracklist/playlist"},
    )
    track_id = fields.Integer(
        missing=None, metadata={"description": "Track ID in the current tracklist"}
    )
    track_no = fields.Integer(
        missing=None, metadata={"description": "Track number in the album"}
    )
    date = fields.String(missing=None, metadata={"description": "Track release date"})
    genre = fields.String(missing=None, metadata={"description": "Track genre"})
    type = fields.Constant("track", metadata={"description": "Item type"})

    @pre_load
    def parse(self, track: dict, **_):
        from platypush.plugins.music.mopidy import EmptyTrackException

        uri = (track or {}).get("uri", (track or {}).get("track", {}).get("uri"))
        if not uri:
            raise EmptyTrackException("Empty track")

        tlid = track.get("tlid")
        playlist_pos = track.get("playlist_pos")
        if track.get("track"):
            track = track.get("track", {})

        length = track.get("length", track.get("time", track.get("duration")))
        return {
            "uri": uri,
            "artist": next(
                iter(item.get("name") for item in track.get("artists", [])),
                None,
            ),
            "title": track.get("name"),
            "album": track.get("album", {}).get("name"),
            "artist_uri": next(
                iter(item.get("uri") for item in track.get("artists", [])), None
            ),
            "album_uri": track.get("album", {}).get("uri"),
            "time": length / 1000 if length is not None else None,
            "playlist_pos": (
                track.get("playlist_pos") if playlist_pos is None else playlist_pos
            ),
            "date": track.get("date", track.get("album", {}).get("date")),
            "track_id": tlid,
            "track_no": track.get("track_no"),
            "genre": track.get("genre"),
        }

    @post_dump
    def to_dict(self, track: dict, **_):
        """
        Fill/move missing fields in the dictionary.
        """
        return {
            "file": track["uri"],
            **track,
        }


class MopidyStatusSchema(Schema):
    """
    Mopidy status schema.
    """

    state = fields.Enum(
        PlayerState,
        required=True,
        metadata={"description": "Player state"},
    )
    volume = fields.Float(metadata={"description": "Player volume (0-100)"})
    consume = fields.Boolean(metadata={"description": "Consume mode"})
    random = fields.Boolean(metadata={"description": "Random mode"})
    repeat = fields.Boolean(metadata={"description": "Repeat mode"})
    single = fields.Boolean(metadata={"description": "Single mode"})
    mute = fields.Boolean(metadata={"description": "Mute mode"})
    time = fields.Float(metadata={"description": "Current time (in seconds)"})
    playing_pos = fields.Integer(
        metadata={"description": "Index of the currently playing track"}
    )
    track = fields.Nested(
        MopidyTrackSchema, missing=None, metadata={"description": "Current track"}
    )

    @post_dump
    def post_dump(self, data: dict, **_):
        """
        Post-dump hook.
        """
        state = data.get("state")
        if state:
            data["state"] = getattr(PlayerState, state).value
        return data


class MopidyPlaylistSchema(Schema):
    """
    Mopidy playlist schema.
    """

    # pylint: disable=too-few-public-methods
    class Meta:  # type: ignore
        """
        Mopidy playlist schema metadata.
        """

        unknown = EXCLUDE

    uri = fields.String(required=True, metadata={"description": "Playlist URI"})
    name = fields.String(required=True, metadata={"description": "Playlist name"})
    last_modified = DateTime(metadata={"description": "Last modified timestamp"})
    tracks = fields.List(
        fields.Nested(MopidyTrackSchema),
        missing=None,
        metadata={"description": "Playlist tracks"},
    )
    type = fields.Constant("playlist", metadata={"description": "Item type"})

    @pre_dump
    def pre_dump(self, playlist, **_):
        """
        Pre-dump hook.
        """
        last_modified = (
            playlist.last_modified
            if hasattr(playlist, "last_modified")
            else playlist.get("last_modified")
        )

        if last_modified:
            last_modified /= 1000
        if hasattr(playlist, "last_modified"):
            playlist.last_modified = last_modified
        else:
            playlist["last_modified"] = last_modified

        return playlist


class MopidyArtistSchema(Schema):
    """
    Mopidy artist schema.
    """

    uri = fields.String(required=True, metadata={"description": "Artist URI"})
    file = fields.String(
        metadata={"description": "Artist URI, for MPD compatibility purposes"}
    )
    name = fields.String(missing=None, metadata={"description": "Artist name"})
    artist = fields.String(
        missing=None,
        metadata={"description": "Same as name - for MPD compatibility purposes"},
    )
    type = fields.Constant("artist", metadata={"description": "Item type"})

    @post_dump
    def to_dict(self, artist: dict, **_):
        """
        Fill/move missing fields in the dictionary.
        """
        return {
            "file": artist["uri"],
            "artist": artist["name"],
            **artist,
        }


class MopidyAlbumSchema(Schema):
    """
    Mopidy album schema.
    """

    uri = fields.String(required=True, metadata={"description": "Album URI"})
    file = fields.String(
        metadata={"description": "Artist URI, for MPD compatibility purposes"}
    )
    artist = fields.String(missing=None, metadata={"description": "Artist name"})
    album = fields.String(
        missing=None,
        metadata={"description": "Same as name - for MPD compatibility purposes"},
    )
    name = fields.String(missing=None, metadata={"description": "Album name"})
    artist_uri = fields.String(missing=None, metadata={"description": "Artist URI"})
    date = fields.String(missing=None, metadata={"description": "Album release date"})
    genre = fields.String(missing=None, metadata={"description": "Album genre"})

    def parse(self, data: dict, **_):
        assert data.get("uri"), "Album URI is required"
        return {
            "uri": data["uri"],
            "artist": data.get("artist")
            or next(
                iter(item.get("name") for item in data.get("artists", [])),
                None,
            ),
            "name": data.get("name"),
            "artist_uri": data.get("artist_uri")
            or next(iter(item.get("uri") for item in data.get("artists", [])), None),
            "album_uri": data.get("album_uri") or data.get("album", {}).get("uri"),
            "date": data.get("date", data.get("album", {}).get("date")),
            "genre": data.get("genre"),
        }

    @pre_load
    def pre_load(self, album: dict, **_):
        """
        Pre-load hook.
        """
        return self.parse(album)

    @pre_dump
    def pre_dump(self, album: dict, **_):
        """
        Pre-dump hook.
        """
        return self.parse(album)

    @post_dump
    def to_dict(self, album: dict, **_):
        """
        Fill/move missing fields in the dictionary.
        """
        return {
            "file": album["uri"],
            "album": album["name"],
            **album,
        }


class MopidyDirectorySchema(Schema):
    """
    Mopidy directory schema.
    """

    uri = fields.String(required=True, metadata={"description": "Directory URI"})
    name = fields.String(required=True, metadata={"description": "Directory name"})
    type = fields.Constant("directory", metadata={"description": "Item type"})


class MopidyFilterSchema(Schema):
    """
    Mopidy filter schema.
    """

    uris = fields.List(fields.String, metadata={"description": "Filter by URIs"})
    artist = fields.List(fields.String, metadata={"description": "Artist name(s)"})
    album = fields.List(fields.String, metadata={"description": "Album name(s)"})
    title = fields.List(fields.String, metadata={"description": "Track title(s)"})
    albumartist = fields.List(
        fields.String, metadata={"description": "Album artist name(s)"}
    )
    date = fields.List(fields.String, metadata={"description": "Track release date(s)"})
    genre = fields.List(fields.String, metadata={"description": "Genre(s)"})
    comment = fields.List(fields.String, metadata={"description": "Comment(s)"})
    disc_no = fields.List(fields.Integer, metadata={"description": "Disc number(s)"})
    musicbrainz_artistid = fields.List(
        fields.String, metadata={"description": "MusicBrainz artist ID(s)"}
    )
    musicbrainz_albumid = fields.List(
        fields.String, metadata={"description": "MusicBrainz album ID(s)"}
    )
    musicbrainz_trackid = fields.List(
        fields.String, metadata={"description": "MusicBrainz album artist ID(s)"}
    )
    any = fields.List(
        fields.String, metadata={"description": "Generic search string(s)"}
    )

    @pre_load
    def pre_load(self, data: dict, **_):
        """
        Pre-load hook.
        """
        for field_name, field in self.fields.items():
            value = data.get(field_name)

            # Back-compatibtility with MPD's single-value filters
            if (
                value is not None
                and isinstance(field, fields.List)
                and isinstance(value, str)
            ):
                data[field_name] = [value]

        return data

    @post_load
    def post_load(self, data: dict, **_):
        """
        Post-load hook.
        """
        title = data.pop("title", None)
        if title:
            data["track_name"] = title

        return data



from dataclasses import dataclass, field
from typing import List, Optional
from datetime import datetime

from platypush.schemas.torrent import TorrentResultSchema


@dataclass
class TorrentSearchResult:
    """
    Data class for results returned by
    :meth:`platypush.plugins.torrent.TorrentPlugin.search`.
    """

    title: str
    file: Optional[str] = None
    url: Optional[str] = None
    provider: Optional[str] = None
    type: Optional[str] = None
    size: int = 0
    duration: float = 0
    language: Optional[str] = None
    category: Optional[str] = None
    seeds: int = 0
    peers: int = 0
    image: Optional[str] = None
    description: Optional[str] = None
    is_media: bool = False
    imdb_id: Optional[str] = None
    tvdb_id: Optional[str] = None
    year: Optional[int] = None
    created_at: Optional[datetime] = None
    quality: Optional[str] = None
    overview: Optional[str] = None
    trailer: Optional[str] = None
    genres: List[str] = field(default_factory=list)
    rating: Optional[float] = None
    critic_rating: Optional[float] = None
    community_rating: Optional[float] = None
    votes: Optional[int] = None
    series: Optional[str] = None
    season: Optional[int] = None
    episode: Optional[int] = None
    num_seasons: Optional[int] = None
    country: Optional[str] = None
    network: Optional[str] = None
    series_status: Optional[str] = None

    def to_dict(self) -> dict:
        return dict(TorrentResultSchema().dump(self))



from sqlalchemy import Column, Integer, ForeignKey

from platypush.common.db import is_defined

from .dimmers import Dimmer
from .switches import Switch


if not is_defined('volume'):

    class Volume(Dimmer):
        __tablename__ = 'volume'

        id = Column(
            Integer, ForeignKey(Dimmer.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('muted'):

    class Muted(Switch):
        __tablename__ = 'muted'

        id = Column(
            Integer, ForeignKey(Switch.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from ._base import XmppBaseHandler
from ._connection import XmppConnectionHandler
from ._conversation import XmppConversationHandler
from ._discover import discover_handlers
from ._message import XmppMessageHandler
from ._ping import XmppPingHandler
from ._presence import XmppPresenceHandler
from ._pubsub import XmppPubSubHandler
from ._registry import XmppHandlersRegistry
from ._room import XmppRoomHandler
from ._roster import XmppRosterHandler


__all__ = [
    "XmppBaseHandler",
    "XmppConnectionHandler",
    "XmppConversationHandler",
    "XmppHandlersRegistry",
    "XmppMessageHandler",
    "XmppPingHandler",
    "XmppPresenceHandler",
    "XmppPubSubHandler",
    "XmppRoomHandler",
    "XmppRosterHandler",
    "discover_handlers",
]



"""Base alembic version

Revision ID: d030953a871d
Revises:
Create Date: 2023-04-28 22:32:49.460118

"""


# revision identifiers, used by Alembic.
revision = 'd030953a871d'
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    pass


def downgrade() -> None:
    pass



from ._classes import MajorDeviceClass, MajorServiceClass, MinorDeviceClass
from ._protocol import Protocol
from ._service import ServiceClass

__all__ = [
    "MajorDeviceClass",
    "MajorServiceClass",
    "MinorDeviceClass",
    "Protocol",
    "ServiceClass",
]



from abc import ABC, abstractmethod
from contextlib import suppress
from enum import Enum
import json
import logging
import signal
import subprocess
import threading
import time
from typing import Any, Callable, Optional, Type

import requests

from platypush.message.event.media import (
    MediaDownloadCancelledEvent,
    MediaDownloadClearEvent,
    MediaDownloadCompletedEvent,
    MediaDownloadErrorEvent,
    MediaDownloadEvent,
    MediaDownloadPausedEvent,
    MediaDownloadProgressEvent,
    MediaDownloadResumedEvent,
    MediaDownloadStartedEvent,
)

from platypush.utils import wait_for_either


class DownloadState(Enum):
    """
    Enum that represents the status of a download.
    """

    IDLE = 'idle'
    STARTED = 'started'
    DOWNLOADING = 'downloading'
    PAUSED = 'paused'
    COMPLETED = 'completed'
    CANCELLED = 'cancelled'
    ERROR = 'error'


class DownloadThread(threading.Thread, ABC):
    """
    Thread that downloads a URL to a file.
    """

    _progress_update_interval = 1
    """ Throttle the progress updates to this interval, in seconds. """

    def __init__(
        self,
        path: str,
        url: str,
        post_event: Callable,
        size: Optional[int] = None,
        timeout: Optional[int] = 10,
        on_start: Callable[['DownloadThread'], None] = lambda _: None,
        on_close: Callable[['DownloadThread'], None] = lambda _: None,
        stop_event: Optional[threading.Event] = None,
    ):
        super().__init__(name=f'DownloadThread-{path}')
        self.path = path
        self.url = url
        self.size = size
        self.timeout = timeout
        self.state = DownloadState.IDLE
        self.progress = None
        self.started_at = None
        self.ended_at = None
        self._upstream_stop_event = stop_event or threading.Event()
        self._stop_event = threading.Event()
        self._post_event = post_event
        self._on_start = on_start
        self._on_close = on_close
        self._paused = threading.Event()
        self._downloading = threading.Event()
        self._last_progress_update_time = 0
        self.logger = logging.getLogger(__name__)

    def should_stop(self) -> bool:
        return self._stop_event.is_set() or self._upstream_stop_event.is_set()

    @abstractmethod
    def _run(self) -> bool:
        pass

    def pause(self):
        self.state = DownloadState.PAUSED
        self._paused.set()
        self._downloading.clear()
        self.post_event(MediaDownloadPausedEvent)

    def resume(self):
        self.state = DownloadState.DOWNLOADING
        self._paused.clear()
        self._downloading.set()
        self.post_event(MediaDownloadResumedEvent)

    def run(self):
        super().run()
        interrupted = False

        try:
            self.on_start()
            interrupted = not self._run()

            if interrupted:
                self.state = DownloadState.CANCELLED
            else:
                self.state = DownloadState.COMPLETED
        except Exception as e:
            self.state = DownloadState.ERROR
            self.post_event(MediaDownloadErrorEvent, error=str(e))
            self.logger.warning('Error while downloading URL: %s', e)
        finally:
            self.on_close()

    def stop(self):
        self.state = DownloadState.CANCELLED
        self._stop_event.set()
        self._downloading.clear()

    def on_start(self):
        self.state = DownloadState.STARTED
        self.started_at = time.time()
        self.post_event(MediaDownloadStartedEvent)
        self._on_start(self)

    def on_close(self):
        self.ended_at = time.time()
        if self.state == DownloadState.CANCELLED:
            self.post_event(MediaDownloadCancelledEvent)
        elif self.state == DownloadState.COMPLETED:
            self.post_event(MediaDownloadCompletedEvent)

        self._on_close(self)

    def clear(self):
        if self.state not in (DownloadState.COMPLETED, DownloadState.CANCELLED):
            self.logger.info(
                'Download thread for %s is still active, stopping', self.url
            )

            self.stop()
            self.join(timeout=10)

        self.post_event(MediaDownloadClearEvent)

    def post_event(self, event_type: Type[MediaDownloadEvent], **kwargs):
        kwargs = {
            'resource': self.url,
            'path': self.path,
            'state': self.state.value,
            'size': self.size,
            'timeout': self.timeout,
            'progress': self.progress,
            'started_at': self.started_at,
            'ended_at': self.ended_at,
            **kwargs,
        }

        self._post_event(event_type, **kwargs)

    def __setattr__(self, name: str, value: Optional[Any], /) -> None:
        if name == 'progress' and value is not None:
            if value < 0 or value > 100:
                self.logger.debug('Invalid progress value:%s', value)
                return

            prev_progress = getattr(self, 'progress', None)

            if prev_progress is None or (
                int(prev_progress) != int(value)
                and (
                    time.time() - self._last_progress_update_time
                    >= self._progress_update_interval
                )
            ):
                value = round(value, 2)
                self._last_progress_update_time = time.time()
                self.post_event(MediaDownloadProgressEvent, progress=value)

        super().__setattr__(name, value)


class FileDownloadThread(DownloadThread):
    """
    Thread that downloads a generic URL to a file.
    """

    def _run(self):
        interrupted = False

        with requests.get(self.url, timeout=self.timeout, stream=True) as response:
            response.raise_for_status()
            self.size = int(response.headers.get('Content-Length', 0)) or None

            with open(self.path, 'wb') as f:
                self.on_start()

                for chunk in response.iter_content(chunk_size=8192):
                    if not chunk or self.should_stop():
                        interrupted = self.should_stop()
                        if interrupted:
                            self.stop()

                        break

                    self.state = DownloadState.DOWNLOADING
                    f.write(chunk)
                    percent = f.tell() / self.size * 100 if self.size else 0
                    self.progress = percent

                    if self._paused.is_set():
                        wait_for_either(self._downloading, self._stop_event)

        return not interrupted


class YouTubeDownloadThread(DownloadThread):
    """
    Thread that downloads a YouTube URL to a file.
    """

    def __init__(
        self,
        *args,
        ytdl: str,
        youtube_format: Optional[str] = None,
        only_audio: bool = False,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._ytdl = ytdl
        self._youtube_format = youtube_format
        self._only_audio = only_audio
        self._proc = None
        self._proc_lock = threading.Lock()

    def _parse_progress(self, line: str):
        try:
            progress = json.loads(line)
        except json.JSONDecodeError:
            return

        status = progress.get('status')
        if not status:
            return

        if status == 'finished':
            self.progress = 100
            return

        if status == 'paused':
            self.state = DownloadState.PAUSED
        elif status == 'downloading':
            self.state = DownloadState.DOWNLOADING

        self.size = int(progress.get('total_bytes_estimate', 0)) or self.size
        if self.size:
            downloaded = int(progress.get('downloaded_bytes', 0))
            self.progress = (downloaded / self.size) * 100

    def _run(self):
        ytdl_cmd = [
            self._ytdl,
            '--newline',
            '--progress',
            '--progress-delta',
            str(self._progress_update_interval),
            '--progress-template',
            '%(progress)j',
            *(['-x'] if self._only_audio else []),
            *(['-f', self._youtube_format] if self._youtube_format else []),
            self.url,
            '-o',
            self.path,
        ]

        self.logger.info('Executing command %r', ytdl_cmd)
        err = None

        with subprocess.Popen(ytdl_cmd, stdout=subprocess.PIPE) as self._proc:
            if self._proc.stdout:
                for line in self._proc.stdout:
                    self.logger.debug(
                        '%s output: %s', self._ytdl, line.decode().strip()
                    )

                    self._parse_progress(line.decode())

                    if self.should_stop():
                        self.stop()
                        return self._proc.returncode == 0

                    if self._paused.is_set():
                        wait_for_either(self._downloading, self._stop_event)

        if self._proc.returncode != 0:
            err = self._proc.stderr.read().decode() if self._proc.stderr else None
            raise RuntimeError(
                f'{self._ytdl} failed with return code {self._proc.returncode}: {err}'
            )

        return True

    def pause(self):
        with self._proc_lock:
            if self._proc:
                self._proc.send_signal(signal.SIGSTOP)

        super().pause()

    def resume(self):
        with self._proc_lock:
            if self._proc:
                self._proc.send_signal(signal.SIGCONT)

        super().resume()

    def stop(self):
        state = None

        with suppress(IOError, OSError), self._proc_lock:
            if self._proc:
                if self._proc.poll() is None:
                    self._proc.terminate()
                    self._proc.wait(timeout=3)
                    if self._proc.returncode is None:
                        self._proc.kill()

                    state = DownloadState.CANCELLED
                elif self._proc.returncode != 0:
                    state = DownloadState.ERROR
                else:
                    state = DownloadState.COMPLETED

                self._proc = None

        super().stop()

        if state:
            self.state = state

    def on_close(self):
        self.stop()
        super().on_close()



import base64
import email
import logging
import json
from email.message import Message

from collections import defaultdict
from datetime import datetime
from enum import Enum
from typing import Any, Dict, IO, List, Optional, Union

from platypush.message import JSONAble

logger = logging.getLogger(__name__)

_text_types = {
    'text/plain',
    'text/html',
    'text/rtf',
    'text/enriched',
    'text/markdown',
    'application/rtf',
}


class Mail(JSONAble):  # pylint: disable=too-few-public-methods
    """
    Model for a mail message.
    """

    def __init__(
        self,
        id: int,  # pylint: disable=redefined-builtin
        date: datetime,
        size: int,
        from_: Optional[Union[Dict[str, str], List[str]]] = None,
        to: Optional[Union[Dict[str, str], List[str]]] = None,
        cc: Optional[Union[Dict[str, str], List[str]]] = None,
        bcc: Optional[Union[Dict[str, str], List[str]]] = None,
        subject: str = '',
        content: Optional[Any] = None,
        **kwargs,
    ):
        self.id = id
        self.date = date
        self.size = size
        self.from_ = from_ or kwargs.pop('from', None)
        self.to = to
        self.cc = cc or []
        self.bcc = bcc or []
        self.subject = subject
        self._content = content
        self.args = kwargs

    @staticmethod
    def _parse_body(msg: Message) -> str:
        body = ''
        if msg.is_multipart():
            for part in msg.walk():
                if (
                    part.get_content_type() in _text_types
                    and
                    # skip any text/plain (txt) attachments
                    'attachment' not in part.get('Content-Disposition', '')
                ):
                    body = bytes(part.get_payload(decode=True)).decode()
                    break
        else:
            body = bytes(msg.get_payload(decode=True)).decode()

        return body

    @staticmethod
    def _parse_attachments(msg: Message) -> List[dict]:
        attachments = []
        if msg.is_multipart():
            for part in msg.walk():
                if 'attachment' not in part.get('Content-Disposition', ''):
                    continue

                raw_payload = bytes(part.get_payload(decode=True))
                try:
                    # Try to decode it as a string
                    payload = raw_payload.decode()
                except UnicodeDecodeError:
                    # Otherwise, return a JSON-encoded string
                    payload = base64.b64encode(raw_payload).decode()

                attachments.append(
                    {
                        'filename': part.get_filename(),
                        'headers': dict(part),
                        'body': payload,
                    }
                )

        return attachments

    @classmethod
    def _parse_payload(cls, payload: Union[bytes, bytearray]) -> dict:
        msg = email.message_from_bytes(payload)
        return {
            'headers': dict(msg),
            'body': cls._parse_body(msg),
            'attachments': cls._parse_attachments(msg),
        }

    @property
    def content(self):
        if isinstance(self._content, (bytes, bytearray)):
            try:
                return self._parse_payload(self._content)
            except Exception as e:
                logger.warning(
                    'Error while parsing payload for message ID %s: %s', self.id, e
                )
                logger.exception(e)

        return self._content

    def to_json(self) -> dict:
        return {
            'id': self.id,
            'date': self.date,
            'size': self.size,
            'from': self.from_,
            'to': self.to,
            'cc': self.cc,
            'bcc': self.bcc,
            'subject': self.subject,
            'content': self.content,
            **self.args,
        }


class MailFlagType(Enum):
    """
    Types of supported mail flags.
    """

    FLAGGED = 'flagged'
    UNREAD = 'unread'


FolderStatus = Dict[MailFlagType, Dict[int, Mail]]
FoldersStatus = Dict[str, FolderStatus]


class AccountsStatus:
    """
    Models the status of all the monitored mailboxes.
    """

    def __init__(self):
        self._dict: Dict[str, FoldersStatus] = defaultdict(
            lambda: defaultdict(lambda: {evt: {} for evt in MailFlagType})
        )

    class Serializer(json.JSONEncoder):
        def default(self, o):
            if isinstance(o, datetime):
                return o.isoformat()
            if isinstance(o, MailFlagType):
                return o.value
            return json.JSONEncoder.default(self, o)

    class Deserializer(json.JSONDecoder):
        def __init__(self):
            super().__init__(object_hook=self._hook)

        @staticmethod
        def _hook(o):
            if 'date' in o:
                o['date'] = datetime.fromisoformat(o['date'])
            if 'flag' in o:
                o['flag'] = MailFlagType(o['flag'])
            return o

    def __getitem__(self, item: str) -> FoldersStatus:
        return self._dict[item]

    def __setitem__(self, key: str, value: FoldersStatus):
        self._dict[key] = value

    def __delitem__(self, key: str) -> None:
        del self._dict[key]

    def __iter__(self):
        return iter(self._dict)

    def __len__(self) -> int:
        return len(self._dict)

    def __contains__(self, item) -> bool:
        return item in self._dict

    def __str__(self):
        return json.dumps(self._dict, cls=self.Serializer)

    def __repr__(self):
        return self.__str__()

    def items(self):
        return self._dict.items()

    def copy(self) -> 'AccountsStatus':
        obj = AccountsStatus()
        obj._dict.update(
            {
                account: {
                    folder: {MailFlagType(evt): msgs for evt, msgs in statuses.items()}
                    for folder, statuses in folders.items()
                }
                for account, folders in self._dict.items()
            }
        )
        return obj

    def get(self, key: str, default=None) -> Optional[FoldersStatus]:
        return self._dict.get(key, default)

    @classmethod
    def read(cls, f: IO) -> 'AccountsStatus':
        obj = cls()
        obj._dict.update(
            {
                account: {
                    folder: {MailFlagType(evt): msgs for evt, msgs in statuses.items()}
                    for folder, statuses in folders.items()
                }
                for account, folders in json.load(f, cls=cls.Deserializer).items()
            }
        )
        return obj

    def write(self, f: IO):
        f.write(
            json.dumps(
                {
                    account: {
                        folder: {
                            evt.value: {msg_id: {} for msg_id in msgs}
                            for evt, msgs in statuses.items()
                        }
                        for folder, statuses in folders.items()
                    }
                    for account, folders in self._dict.items()
                },
                cls=self.Serializer,
            )
        )


# vim:sw=4:ts=4:et:



from typing import (
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Union,
)

from pyHS100 import (
    SmartDevice,
    SmartPlug,
    SmartBulb,
    SmartStrip,
    Discover,
    SmartDeviceException,
)

from platypush.entities import Entity, SwitchEntityManager
from platypush.plugins import RunnablePlugin, action


class SwitchTplinkPlugin(RunnablePlugin, SwitchEntityManager):
    """
    Plugin to interact with TP-Link smart switches/plugs like the HS100
    (https://www.tp-link.com/us/products/details/cat-5516_HS100.html).
    """

    _ip_to_dev: Dict[str, SmartDevice] = {}
    _alias_to_dev: Dict[str, SmartDevice] = {}

    def __init__(
        self,
        plugs: Optional[Union[Mapping[str, str], List[str]]] = None,
        bulbs: Optional[Union[Mapping[str, str], List[str]]] = None,
        strips: Optional[Union[Mapping[str, str], List[str]]] = None,
        **kwargs,
    ):
        """
        :param plugs: Optional list of IP addresses or name->address mapping if you have a static list of
            TpLink plugs and you want to save on the scan time.
        :param bulbs: Optional list of IP addresses or name->address mapping if you have a static list of
            TpLink bulbs and you want to save on the scan time.
        :param strips: Optional list of IP addresses or name->address mapping if you have a static list of
            TpLink strips and you want to save on the scan time.
        """
        super().__init__(**kwargs)
        self._ip_to_dev = {}
        self._alias_to_dev = {}
        self._static_devices = {}

        if isinstance(plugs, list):
            plugs = {addr: addr for addr in plugs}
        if isinstance(bulbs, list):
            bulbs = {addr: addr for addr in bulbs}
        if isinstance(strips, list):
            strips = {addr: addr for addr in strips}

        for name, addr in (plugs or {}).items():
            self._static_devices[addr] = {
                'name': name,
                'type': SmartPlug,
            }

        for name, addr in (bulbs or {}).items():
            self._static_devices[addr] = {
                'name': name,
                'type': SmartBulb,
            }

        for name, addr in (strips or {}).items():
            self._static_devices[addr] = {
                'name': name,
                'type': SmartStrip,
            }

        self._update_devices()

    def _update_devices(
        self,
        devices: Optional[Mapping[str, SmartDevice]] = None,
        publish_entities: bool = True,
    ):
        for addr, info in self._static_devices.items():
            try:
                dev = info['type'](addr)
                self._alias_to_dev[info.get('name', dev.alias)] = dev
                self._ip_to_dev[addr] = dev
            except SmartDeviceException as e:
                self.logger.warning('Could not communicate with device %s: %s', addr, e)

        for ip, dev in (devices or {}).items():
            self._ip_to_dev[ip] = dev
            self._alias_to_dev[dev.alias] = dev

        if devices and publish_entities:
            self.publish_entities(devices.values())

    def transform_entities(self, entities: Collection[SmartDevice]):
        from platypush.entities.switches import Switch

        return super().transform_entities(
            [
                Switch(
                    id=dev.host,
                    name=dev.alias,
                    state=dev.is_on,
                    data={
                        'current_consumption': dev.current_consumption(),
                        'ip': dev.host,
                        'host': dev.host,
                        'hw_info': dev.hw_info,
                    },
                )
                for dev in (entities or [])
            ]
        )

    def _scan(self, publish_entities: bool = True) -> Dict[str, SmartDevice]:
        devices = Discover.discover()
        self._update_devices(devices, publish_entities=publish_entities)
        return devices

    def _get_device(self, device, use_cache=True):
        if not use_cache:
            self._scan()

        if isinstance(device, Entity):
            device = device.external_id or device.name

        if device in self._ip_to_dev:
            return self._ip_to_dev[device]

        if device in self._alias_to_dev:
            return self._alias_to_dev[device]

        if use_cache:
            return self._get_device(device, use_cache=False)
        raise RuntimeError(f'Device {device} not found')

    def _set(self, device: SmartDevice, state: bool):
        action_name = 'turn_on' if state else 'turn_off'
        act = getattr(device, action_name, None)
        assert act, (
            f'No such action available on the device "{device.alias}": '
            f'"{action_name}"'
        )
        act()
        self.publish_entities([device])
        return self._serialize(device)

    @action
    def on(self, device, **_):  # pylint: disable=arguments-differ
        """
        Turn on a device

        :param device: Device IP, hostname or alias
        :type device: str
        """

        device = self._get_device(device)
        return self._set(device, True)

    @action
    def off(self, device, **_):  # pylint: disable=arguments-differ
        """
        Turn off a device

        :param device: Device IP, hostname or alias
        :type device: str
        """

        device = self._get_device(device)
        return self._set(device, False)

    @action
    def toggle(self, device, **_):  # pylint: disable=arguments-differ
        """
        Toggle the state of a device (on/off)

        :param device: Device IP, hostname or alias
        :type device: str
        """

        device = self._get_device(device)
        return self._set(device, not device.is_on)

    @staticmethod
    def _serialize(device: SmartDevice) -> dict:
        return {
            'current_consumption': device.current_consumption(),
            'id': device.host,
            'ip': device.host,
            'host': device.host,
            'hw_info': device.hw_info,
            'name': device.alias,
            'on': device.is_on,
        }

    @action
    def status(self, *_, **__) -> List[dict]:
        """
        Retrieve the current status of the devices. Return format:

            .. code-block:: json

                [
                    {
                        "current_consumption": 0.5,
                        "id": "192.168.1.123",
                        "ip": "192.168.1.123",
                        "host": "192.168.1.123",
                        "hw_info": "00:11:22:33:44:55",
                        "name": "My Switch",
                        "on": true,
                    }
                ]

        """
        return [self._serialize(dev) for dev in self._scan().values()]

    def main(self):
        devices = {ip: self._serialize(dev) for ip, dev in self._ip_to_dev.items()}

        while not self.should_stop():
            new_devices = self._scan(publish_entities=False)
            new_serialized_devices = {
                ip: self._serialize(dev) for ip, dev in new_devices.items()
            }

            updated_devices = {
                ip: new_devices[ip]
                for ip, dev in new_serialized_devices.items()
                if any(v != devices.get(ip, {}).get(k) for k, v in dev.items())
            }

            if updated_devices:
                self.publish_entities(updated_devices.values())

            devices = new_serialized_devices
            self.wait_stop(self.poll_interval)


# vim:sw=4:ts=4:et:



from threading import Thread, current_thread
from typing import Set

from platypush.backend.http.app.utils import send_message
from platypush.message.request import Request

from . import WSRoute, logger


class WSRequestsProxy(WSRoute):
    """
    Websocket event proxy mapped to ``/ws/requests``.
    """

    _max_concurrent_requests: int = 10
    """ Maximum number of concurrent requests allowed on the same connection. """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._requests: Set[Thread] = set()

    @classmethod
    def app_name(cls) -> str:
        return 'requests'

    def _handle_request(self, request: Request):
        self._requests.add(current_thread())
        try:
            response = send_message(request, wait_for_response=True)
            self.send(str(response))
        finally:
            self._requests.remove(current_thread())

    def on_message(self, message):
        if len(self._requests) > self._max_concurrent_requests:
            logger.info('Too many concurrent requests on %s', self)
            return

        try:
            msg = Request.build(message)
            assert isinstance(msg, Request), f'Expected {Request}, got {type(msg)}'
        except Exception as e:
            logger.info('Could not build request from %s: %s', message, e)
            logger.exception(e)
            return

        Thread(target=self._handle_request, args=(msg,)).start()



from platypush.message.event import Event


class FileSystemEvent(Event):
    """
    Base class for file system events - namely, file/directory creation, deletion and modification.
    """
    def __init__(self, path: str, *, is_directory: bool, **kwargs):
        super().__init__(path=path, is_directory=is_directory, **kwargs)


class FileSystemCreateEvent(FileSystemEvent):
    """
    Event triggered when a monitored file or directory is created.
    """


class FileSystemDeleteEvent(FileSystemEvent):
    """
    Event triggered when a monitored file or directory is deleted.
    """


class FileSystemModifyEvent(FileSystemEvent):
    """
    Event triggered when a monitored file or directory is modified.
    """



from dataclasses import asdict, dataclass
from typing import Optional

from platypush.plugins.media import PlayerState
from platypush.schemas.mopidy import MopidyStatusSchema

from ._track import MopidyTrack


@dataclass
class MopidyStatus:
    """
    A dataclass to hold the status of the Mopidy client.
    """

    state: PlayerState = PlayerState.STOP
    volume: float = 0
    consume: bool = False
    random: bool = False
    repeat: bool = False
    single: bool = False
    mute: bool = False
    time: Optional[float] = None
    duration: Optional[float] = None
    playing_pos: Optional[int] = None
    track: Optional[MopidyTrack] = None

    def copy(self):
        return MopidyStatus(
            state=self.state,
            volume=self.volume,
            consume=self.consume,
            random=self.random,
            repeat=self.repeat,
            single=self.single,
            mute=self.mute,
            time=self.time,
            duration=self.duration,
            playing_pos=self.playing_pos,
            track=MopidyTrack(**asdict(self.track)) if self.track else None,
        )

    def to_dict(self):
        """
        Convert the Mopidy status to a dictionary.
        """
        return dict(MopidyStatusSchema().dump(self))



import re
import textwrap as tw
from contextlib import contextmanager
from typing import Callable, Dict, Generator, Optional

from .._model.argument import Argument
from .._model.returns import ReturnValue
from .._serialize import Serializable
from .context import ParseContext
from .rst import RstExtensionsMixin
from .state import ParseState


class DocstringParser(Serializable, RstExtensionsMixin):
    """
    Mixin for objects that can parse docstrings.
    """

    _param_doc_re = re.compile(r"^:param\s+(?P<name>[\w_]+):\s+(?P<doc>.*)$")
    _type_doc_re = re.compile(r"^:type\s+[\w_]+:.*$")
    _return_doc_re = re.compile(r"^:return:\s+(?P<doc>.*)$")
    _default_docstring = re.compile(r"^\s*Initialize self\.\s*See help.*$")

    def __init__(
        self,
        name: str,
        doc: Optional[str] = None,
        args: Optional[Dict[str, Argument]] = None,
        has_varargs: bool = False,
        has_kwargs: bool = False,
        returns: Optional[ReturnValue] = None,
    ):
        self.name = name
        self.doc = doc
        self.args = args or {}
        self.has_varargs = has_varargs
        self.has_kwargs = has_kwargs
        self.returns = returns

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "doc": self.doc,
            "args": {k: v.to_dict() for k, v in self.args.items()},
            "has_varargs": self.has_varargs,
            "has_kwargs": self.has_kwargs,
            "returns": self.returns.to_dict() if self.returns else None,
        }

    @staticmethod
    def _norm_indent(text: Optional[str]) -> Optional[str]:
        """
        Normalize the indentation of a docstring.

        :param text: Input docstring
        :return: A representation of the docstring where all the leading spaces have been removed.
        """
        if not text:
            return None

        lines = text.split("\n")
        return (lines[0] + "\n" + tw.dedent("\n".join(lines[1:]) or "")).strip()

    @classmethod
    @contextmanager
    def _parser(cls, obj: Callable) -> Generator[ParseContext, None, None]:
        """
        Manages the parsing context manager.

        :param obj: Method to parse.
        :return: The parsing context.
        """

        ctx = ParseContext(obj)
        yield ctx

        # Normalize the parameters docstring indentation
        for param in ctx.parsed_params.values():
            param.doc = cls._norm_indent(param.doc)

        # Normalize the return docstring indentation
        ctx.returns.doc = cls._norm_indent(ctx.returns.doc)

    @staticmethod
    def _is_continuation_line(line: str) -> bool:
        return not line.strip() or line.startswith(" ")

    @classmethod
    def _parse_line(cls, line: str, ctx: ParseContext):
        """
        Parse a single line of the docstring and updates the parse context accordingly.

        :param line: Docstring line.
        :param ctx: Parse context.
        """
        # Ignore old in-doc type hints
        if cls._type_doc_re.match(line) or (
            ctx.state == ParseState.TYPE and cls._is_continuation_line(line)
        ):
            ctx.state = ParseState.TYPE
            return

        # Ignore the default constructor docstring
        if cls._default_docstring.match(line):
            return

        # Expand any custom RST extensions
        line = cls._expand_rst_extensions(line, ctx.obj)

        # Update the return type docstring if required
        m = cls._return_doc_re.match(line)
        if m or (ctx.state == ParseState.RETURN and cls._is_continuation_line(line)):
            ctx.state = ParseState.RETURN
            ctx.returns.doc = ((ctx.returns.doc + "\n") if ctx.returns.doc else "") + (
                m.group("doc") if m else line
            ).rstrip()
            return

        # Initialize the documentation of a parameter on :param: docstring lines
        m = cls._param_doc_re.match(line)
        if m and ctx.parsed_params.get(m.group("name")):
            ctx.state = ParseState.PARAM
            ctx.cur_param = m.group("name")

            # Skip vararg/var keyword parameters
            if ctx.cur_param in {ctx.spec.varkw, ctx.spec.varargs}:
                return

            ctx.parsed_params[ctx.cur_param].doc = m.group("doc")
            return

        # Update the current parameter docstring if required
        if ctx.state == ParseState.PARAM and cls._is_continuation_line(line):
            if ctx.cur_param in ctx.parsed_params:
                ctx.parsed_params[ctx.cur_param].doc = (
                    (
                        (ctx.parsed_params[ctx.cur_param].doc or "")
                        + "\n"
                        + line.rstrip()
                    )
                    if ctx.parsed_params.get(ctx.cur_param)
                    and ctx.parsed_params[ctx.cur_param].doc
                    else ""
                )
            return

        # Update the current docstring if required
        ctx.cur_param = None
        ctx.doc = ((ctx.doc + "\n") if ctx.doc else "") + line.rstrip()
        ctx.state = ParseState.DOC

    @classmethod
    def parse(cls, obj: Callable):
        """
        Parse the parameters of a class constructor or action method.
        :param obj: Method to parse.
        :return: The parsed parameters.
        """
        with cls._parser(obj) as ctx:
            for line in ctx.doc_lines:
                cls._parse_line(line, ctx)

        return cls(
            name=obj.__name__,
            doc=ctx.doc,
            args=ctx.parsed_params,
            has_varargs=ctx.spec.varargs is not None,
            has_kwargs=ctx.spec.varkw is not None,
            returns=ctx.returns,
        )



from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional

from platypush.schemas.mopidy import MopidyPlaylistSchema

from ._track import MopidyTrack


@dataclass
class MopidyPlaylist:
    """
    Model for a Mopidy playlist.
    """

    uri: str
    name: str
    last_modified: Optional[datetime] = None
    tracks: List[MopidyTrack] = field(default_factory=list)
    type: str = "playlist"

    @classmethod
    def parse(cls, playlist: dict) -> "MopidyPlaylist":
        """
        Parse a Mopidy playlist from a dictionary received from the Mopidy API.
        """
        return cls(**MopidyPlaylistSchema().load(playlist))  # type: ignore

    def to_dict(self) -> dict:
        """
        Convert the Mopidy playlist to a dictionary.
        """
        return dict(MopidyPlaylistSchema().dump(self))



from marshmallow_dataclass import class_schema

from platypush.schemas.dataclasses import DataClassSchema

from ._model import SystemInfo


SystemInfoSchema = class_schema(SystemInfo, base_schema=DataClassSchema)



import logging

from platypush.message.event import Event


class DistanceSensorEvent(Event):
    """
    Event triggered when a new value is processed by a distance sensor.
    """

    def __init__(self, distance: float, unit: str = 'mm', *args, **kwargs):
        super().__init__(
            *args,
            distance=distance,
            unit=unit,
            logging_level=logging.DEBUG,
            disable_web_clients_notification=True,
            **kwargs
        )


# vim:sw=4:ts=4:et:



import json

from tornado.web import stream_request_body

from platypush.backend.http.app.streaming import StreamingRoute
from platypush.backend.http.app.utils.bus import send_request

from ._registry import load_media_map


@stream_request_body
class MediaSubtitlesRoute(StreamingRoute):
    """
    Route for media stream subtitles.
    """

    SUPPORTED_METHODS = ['GET', 'POST', 'DELETE']

    @classmethod
    def path(cls) -> str:
        return r"^/media/subtitles/([a-zA-Z0-9_.]+)\.vtt$"

    @property
    def auth_required(self) -> bool:
        return False

    def get(self, media_id: str):
        """
        GET route to retrieve the subtitles for the given media_id.
        """
        try:
            self.get_subtitles(media_id)
        except Exception as e:
            self._on_error(e)

    def post(self, media_id: str):
        """
        POST route to add subtitles to the given media_id.
        """
        try:
            self.add_subtitles(media_id)
        except Exception as e:
            self._on_error(e)

    def delete(self, media_id: str):
        """
        DELETE route to remove the subtitles for the given media_id.
        """
        try:
            self.remove_subtitles(media_id)
        except Exception as e:
            self._on_error(e)

    def get_subtitles(self, media_id: str):
        """
        Retrieves the subtitles for the given media_id.
        """

        media_hndl = load_media_map().get(media_id)
        if not media_hndl:
            raise FileNotFoundError(f'{media_id} is not a registered media_id')

        if not media_hndl.subtitles:
            raise FileNotFoundError(f'{media_id} has no subtitles')

        with open(media_hndl.subtitles) as f:
            self.set_header('Content-Type', 'text/vtt')
            self.finish(f.read())

    @staticmethod
    def remove_subtitles(media_id: str):
        """
        Remove the current subtitle track from a streamed from a media file.
        """
        media_hndl = load_media_map().get(media_id)
        if not media_hndl:
            raise FileNotFoundError(f'{media_id} is not a registered media_id')

        if not media_hndl.subtitles:
            raise FileNotFoundError(f'{media_id} has no subtitles attached')

        media_hndl.remove_subtitles()
        return {}

    def add_subtitles(self, media_id: str):
        """
        This route can be used to download and/or expose subtitles files
        associated to a media file
        """

        media_hndl = load_media_map().get(media_id)
        if not media_hndl:
            raise FileNotFoundError(f'{media_id} is not a registered media_id')

        subfile = None
        if self.request.body:
            subfile = json.loads(self.request.body).get('filename')
            assert subfile, 'No filename specified in the request'

        if not subfile:
            if not media_hndl.path:
                raise NotImplementedError(
                    'Subtitles are currently only supported for local media files'
                )

            req = {
                'type': 'request',
                'action': 'media.subtitles.get_subtitles',
                'args': {
                    'resource': media_hndl.path,
                },
            }

            try:
                subtitles = send_request(req) or []
            except Exception as e:
                raise RuntimeError(f'Could not get subtitles: {e}') from e

            if not subtitles:
                raise FileNotFoundError(
                    f'No subtitles found for resource {media_hndl.path}'
                )

            req = {
                'type': 'request',
                'action': 'media.subtitles.download',
                'args': {
                    'link': subtitles[0].get('SubDownloadLink'),
                    'media_resource': media_hndl.path,
                    'convert_to_vtt': True,
                },
            }

            subfile = (send_request(req) or {}).get('filename')

        media_hndl.set_subtitles(subfile)
        return {
            'filename': subfile,
            'url': f'/media/subtitles/{media_id}.vtt',
        }



import logging
import threading

from typing import Optional

import gi

gi.require_version('Gst', '1.0')
gi.require_version('GstApp', '1.0')

# flake8: noqa
from gi.repository import GLib, Gst, GstApp

Gst.init(None)


class Pipeline:
    """
    A GStreamer pipeline.
    """

    def __init__(self):
        self.logger = logging.getLogger('gst-pipeline')
        self.pipeline = Gst.Pipeline()
        self.pipeline.set_state(Gst.State.NULL)
        self.loop = Loop()
        self.source = None
        self.sink = None

        self.bus = self.pipeline.get_bus()
        self.bus.add_signal_watch()
        self.bus.connect('message::eos', self.on_eos)
        self.bus.connect('message::error', self.on_error)
        self.data_ready = threading.Event()
        self.data = None

    def add(self, element_name: str, *args, **props):
        el = Gst.ElementFactory.make(element_name, *args)
        for k, v in props.items():
            if k == 'caps':
                v = Gst.caps_from_string(v)
            el.set_property(k, v)

        self.pipeline.add(el)
        return el

    def add_source(self, element_name: str, *args, **props):
        assert not self.source, 'A source element is already set for this pipeline'
        source = self.add(element_name, *args, **props)
        self.source = source
        return source

    def add_sink(self, element_name: str, *args, **props):
        assert not self.sink, 'A sink element is already set for this pipeline'
        sink = self.add(element_name, *args, **props)
        sink.connect('new-sample', self.on_buffer)
        sink.set_property('emit-signals', True)
        self.sink = sink
        return sink

    @staticmethod
    def link(*elements):
        for i, el in enumerate(elements):
            if i == len(elements) - 1:
                break
            el.link(elements[i + 1])

    def emit(self, signal, *args, **kwargs):
        return self.pipeline.emit(signal, *args, **kwargs)

    def play(self):
        self.pipeline.set_state(Gst.State.PLAYING)
        assert self.loop, 'No GLib loop is running'
        self.loop.start()

    def pause(self):
        state = self.get_state()
        if state == Gst.State.PAUSED:
            self.pipeline.set_state(Gst.State.PLAYING)
        else:
            self.pipeline.set_state(Gst.State.PAUSED)

    def stop(self):
        self.pipeline.set_state(Gst.State.NULL)
        if self.loop:
            self.loop.stop()
        self.loop = None

    def get_volume(self) -> float:
        assert self.source, 'No source initialized'
        return self.source.get_property('volume') or 0

    def set_volume(self, volume: float):
        assert self.source, 'No source initialized'
        self.source.set_property('volume', volume)

    def on_buffer(self, sink):
        sample = GstApp.AppSink.pull_sample(sink)
        buffer = sample.get_buffer()
        size, offset, _ = buffer.get_sizes()
        self.data = buffer.extract_dup(offset, size)
        self.data_ready.set()
        return False

    def on_eos(self, *_, **__):
        self.logger.info('End of stream event received')
        self.stop()

    def on_error(self, _, msg):
        self.logger.warning('GStreamer pipeline error: %s', msg.parse_error())
        self.stop()

    def get_source(self):
        return self.source

    def get_sink(self):
        return self.sink

    def get_state(self) -> Gst.State:
        if not (self.source and self.source.current_state):
            self.logger.warning('Unable to get pipeline state')
            return Gst.State.NULL

        return self.source.current_state

    def is_playing(self) -> bool:
        return self.get_state() == Gst.State.PLAYING

    def is_paused(self) -> bool:
        return self.get_state() == Gst.State.PAUSED

    def get_position(self) -> Optional[float]:
        if not self.source:
            self.logger.warning('Unable to get pipeline state')
            return Gst.State.NULL

        pos = self.source.query_position(Gst.Format(Gst.Format.TIME))
        if not pos[0]:
            return None

        return pos[1] / 1e9

    def get_duration(self) -> Optional[float]:
        assert self.source, 'No active source found'
        pos = self.source.query_duration(Gst.Format(Gst.Format.TIME))
        if not pos[0]:
            return None

        return pos[1] / 1e9

    def is_muted(self) -> bool:
        if not self.source:
            return False
        return self.source.get_property('mute')

    def set_mute(self, mute: bool):
        assert self.source, 'No source specified'
        self.source.set_property('mute', mute)

    def mute(self):
        self.set_mute(True)

    def unmute(self):
        self.set_mute(False)

    def seek(self, position: float):
        assert self.source, 'No source specified'
        position = max(0, position)
        duration = self.get_duration()
        if duration and position > duration:
            position = duration

        seek_ns = int(position * 1e9)
        self.source.seek_simple(Gst.Format.TIME, Gst.SeekFlags.FLUSH, seek_ns)


class Loop(threading.Thread):
    """
    Wraps the execution of a GLib main loop into its own thread.
    """

    def __init__(self):
        super().__init__()
        self._loop = GLib.MainLoop()

    def run(self):
        assert self._loop, 'No GLib loop is running'
        self._loop.run()

    def is_running(self) -> bool:
        return self.is_alive() or self._loop is not None

    def stop(self):
        if not self.is_running():
            return
        if self._loop:
            self._loop.quit()
        if threading.get_ident() != self.ident:
            self.join()
        self._loop = None


# vim:sw=4:ts=4:et:



import threading
from contextlib import contextmanager
from dataclasses import asdict
from datetime import datetime

from platypush.plugins import RunnablePlugin, action
from platypush.message.event.gps import (
    GPSDeviceEvent,
    GPSDisabledEvent,
    GPSEnabledEvent,
    GPSLocationUpdateEvent,
)
from platypush.schemas.gps import GpsDeviceSchema, GpsStatusSchema

from ._model import DeviceMode, GpsDevice, GpsStatus


class GpsPlugin(RunnablePlugin):
    """
    This plugin can interact with a GPS device compatible with `gpsd
    <https://gpsd.io/>`_ and emit events when the location changes.

    It requires ``gpsd`` to run on a system with a compatible GPS device
    connected - most of the off-the-shelf GPS devices over USB or serial
    interfaces should tick the box.

    For example, if your GPS device communicates over USB and is available on
    /dev/ttyUSB0, you can start the gpsd daemon with the following command
    before starting Platypush::

        [sudo] gpsd /dev/ttyUSB0 [-S 2947]

    It will expose GPS events over the port ``2947`` by default, and you can
    subscribe to them through this plugin.
    """

    _default_gpsd_port = 2947
    _default_poll_interval = 5.0
    _lat_lng_tolerance = 1e-5
    _alt_tolerance = 0.5

    def __init__(
        self,
        gpsd_server: str = 'localhost',
        gpsd_port: int = _default_gpsd_port,
        poll_interval: float = _default_poll_interval,
        enable_on_start: bool = True,
        **kwargs,
    ):
        """
        :param gpsd_server: gpsd daemon server name/address (default: localhost).
        :param gpsd_port: Port of the gpsd daemon (default: 2947).
        :param poll_interval: How long to wait before polling the GPS device
            again in case of error (default: 5 seconds).
        :param enable_on_start: If True, the GPS polling will be enabled when the
            plugin starts (default: True). Otherwise, it'll have to be enabled by
            calling the :meth:`.enable` action.
        """
        super().__init__(poll_interval=poll_interval, **kwargs)

        self.gpsd_server = gpsd_server
        self.gpsd_port = gpsd_port
        self._enable_on_start = enable_on_start
        self._session = None
        self._session_lock = threading.RLock()
        self._status = GpsStatus()

    @contextmanager
    def _get_session(self):
        import gps

        with self._session_lock:
            if not self._session:
                self._session = gps.gps(
                    host=self.gpsd_server, port=self.gpsd_port, reconnect=True  # type: ignore
                )

        yield self._session

        with self._session_lock:
            try:
                self.disable()
            except Exception as e:
                self.logger.warning('Error disabling GPSD watch: %s', e)

            self._session.close()
            self._session = None

    def _update_device(self, device: dict):
        path = device.get('path')
        if not path:
            return

        cur_dev = self._status.devices.get(path)
        new_dev = GpsDevice(**GpsDeviceSchema().load(device))  # type: ignore
        if cur_dev and asdict(cur_dev) == asdict(new_dev):
            return

        self._status.devices[path] = new_dev
        self._bus.post(GPSDeviceEvent(**asdict(new_dev)))

    def _handle_location_update(self, report: dict):
        dev, lat, long, alt, mode, t = (
            report.get('device'),
            report.get('lat'),
            report.get('lon'),
            report.get('alt'),
            report.get('mode'),
            report.get('time'),
        )

        if not (dev and lat and long and mode):
            return

        dev_mode = DeviceMode(mode)
        self._status.timestamp = datetime.fromisoformat(t) if t else None
        self._status.devices[dev].mode = dev_mode

        if not (
            abs((self._status.latitude or 0) - lat) >= self._lat_lng_tolerance
            or abs((self._status.longitude or 0) - long) >= self._lat_lng_tolerance
            or abs((self._status.altitude or 0) - (alt or 0)) >= self._alt_tolerance
        ):
            return

        event = GPSLocationUpdateEvent(
            device=dev,
            latitude=lat,
            longitude=long,
            altitude=alt,
            mode=dev_mode.name,
        )

        self._status.latitude = lat
        self._status.longitude = long
        self._status.altitude = alt
        self._bus.post(event)

    def _handle_report(self, report: dict):
        cls = report['class'].lower()
        if cls == 'version':
            self.logger.info('Received GPSD version event: %s', dict(report))
            return

        if cls == 'watch':
            evt_type = GPSEnabledEvent if report.get('enable') else GPSDisabledEvent
            self._bus.post(evt_type())
            return

        if cls == 'devices':
            for device in report.get('devices', []):
                self._update_device(device)
            return

        if cls == 'device':
            self._update_device(report)
            return

        if cls == 'tpv':
            self._handle_location_update(report)

    @action
    def enable(self):
        """
        Enable the GPS polling.
        """
        import gps

        assert self._session, 'GPSD session not initialized'
        self._session.stream(gps.WATCH_ENABLE | gps.WATCH_NEWSTYLE)

    @action
    def disable(self):
        """
        Disable the GPS polling.
        """
        import gps

        assert self._session, 'GPSD session not initialized'
        self._session.stream(gps.WATCH_DISABLE)

    @action
    def status(self):
        """
        :returns: The current GPS status:

            .. schema:: gps.GpsStatusSchema

        """
        return GpsStatusSchema().dump(self._status)

    def main(self):
        while not self.should_stop():
            first_run = True

            try:
                with self._get_session() as session:
                    if first_run and self._enable_on_start:
                        self.enable()
                        first_run = False

                    while not self.should_stop():
                        report: dict = session.next()  # type: ignore
                        self._handle_report(report)
            except Exception as e:
                if isinstance(e, StopIteration):
                    self.logger.warning(
                        'GPS service connection lost, check that gpsd is running'
                    )
                else:
                    self.logger.exception(e)

                self.wait_stop(self.poll_interval)


# vim:sw=4:ts=4:et:



from typing import List, Optional

from platypush.message.event import Event


class TelegramEvent(Event):
    """
    Base class for all the Telegram events.
    """

    def __init__(self, *args, chat_id: int, **kwargs):
        super().__init__(*args, chat_id=chat_id, **kwargs)


class MessageEvent(TelegramEvent):
    """
    Event triggered when a new message is received by the Telegram bot.
    """

    def __init__(  # pylint: disable=useless-parent-delegation
        self, *args, message: Optional[dict], user: Optional[dict], **kwargs
    ):
        """
        :param message: .. schema:: telegram.TelegramMessageSchema
        :param user: .. schema:: telegram.TelegramUserSchema
        """
        super().__init__(*args, message=message, user=user, **kwargs)


class CommandMessageEvent(MessageEvent):
    """
    Event triggered when a new message is received by the Telegram bot.
    """

    def __init__(  # pylint: disable=useless-parent-delegation
        self, *args, command: str, cmdargs: Optional[List[str]] = None, **kwargs
    ):
        """
        :param command: Command name.
        :param cmdargs: Command arguments.
        """
        super().__init__(*args, command=command, cmdargs=(cmdargs or []), **kwargs)


class TextMessageEvent(MessageEvent):
    """
    Event triggered when a new text message is received by the Telegram bot.
    """


class PhotoMessageEvent(MessageEvent):
    """
    Event triggered when a new photo message is received by the Telegram bot.
    """


class VideoMessageEvent(MessageEvent):
    """
    Event triggered when a new video message is received by the Telegram bot.
    """


class ContactMessageEvent(MessageEvent):
    """
    Event triggered when a new contact message is received by the Telegram bot.
    """


class LocationMessageEvent(MessageEvent):
    """
    Event triggered when a new location message is received by the Telegram bot.
    """


class DocumentMessageEvent(MessageEvent):
    """
    Event triggered when a new document message is received by the Telegram bot.
    """


class GroupChatCreatedEvent(MessageEvent):
    """
    Event triggered when a new group chat is created.
    """


# vim:sw=4:ts=4:et:



from typing import Iterable, Dict, Optional, Union, Any

from watchdog.observers import Observer

from platypush.plugins import RunnablePlugin

from .entities.handlers import EventHandler
from .entities.resources import MonitoredResource, MonitoredPattern, MonitoredRegex


class FileMonitorPlugin(RunnablePlugin):
    """
    A plugin to monitor changes to files and directories.
    """

    def __init__(
        self, paths: Iterable[Union[str, Dict[str, Any], MonitoredResource]], **kwargs
    ):
        """
        :param paths: List of paths to monitor. Paths can either be expressed
            in any of the following ways:

                - Simple strings. In this case, paths will be interpreted as
                  absolute references to a file or a directory to monitor.
                  Example:

                  .. code-block:: yaml

                    file.monitor:
                        paths:
                            # Monitor changes on the /tmp folder
                            - /tmp
                            # Monitor changes on /etc/passwd
                            - /etc/passwd

                - Path with monitoring properties expressed as a key-value
                  object. Example showing the supported attributes:

                    .. code-block:: yaml

                        file.monitor:
                            paths:
                                # Monitor changes on the /tmp folder and its
                                # subfolders
                                - path: /tmp
                                  recursive: True

                - Path with pattern-based search criteria for the files to monitor
                  and exclude. Example:

                    .. code-block:: yaml

                        file.monitor:
                            paths:
                                # Recursively monitor changes on the
                                # ~/my-project folder that include all *.py
                                # files, excluding those whose name starts with
                                # tmp_ and all the files contained in the
                                # __pycache__ folders
                                - path: ~/my-project
                                  recursive: True
                                  patterns:
                                    - "*.py"
                                  ignore_patterns:
                                    - "tmp_*"
                                  ignore_directories:
                                    - "__pycache__"

                - Path with regex-based search criteria for the files to
                  monitor and exclude. Example:

                    .. code-block:: yaml

                        file.monitor:
                            paths:
                                # Recursively monitor changes on the
                                # ~/my-images folder that include all the files
                                # matching a JPEG extension in case-insensitive
                                # mode, excluding those whose name starts with
                                # tmp_ and all the files contained in the
                                # __MACOSX folders
                                - path: ~/my-images
                                  recursive: True
                                  case_sensitive: False
                                  regexes:
                                    - '.*\\.jpe?g$'
                                  ignore_patterns:
                                    - '^tmp_.*'
                                  ignore_directories:
                                    - '__MACOSX'

        """

        super().__init__(**kwargs)
        self._observer = Observer()
        self.paths = set()

        for path in paths:
            handler = self.event_handler_from_resource(path)
            if not handler:
                continue

            self.paths.add(handler.resource.path)
            self._observer.schedule(
                handler, handler.resource.path, recursive=handler.resource.recursive
            )

    @staticmethod
    def event_handler_from_resource(
        resource: Union[str, Dict[str, Any], MonitoredResource]
    ) -> Optional[EventHandler]:
        """
        Create a file system event handler from a string, dictionary or
        ``MonitoredResource`` resource.
        """

        if isinstance(resource, str):
            res = MonitoredResource(resource)
        elif isinstance(resource, dict):
            if 'regexes' in resource or 'ignore_regexes' in resource:
                res = MonitoredRegex(**resource)
            elif (
                'patterns' in resource
                or 'ignore_patterns' in resource
                or 'ignore_directories' in resource
            ):
                res = MonitoredPattern(**resource)
            else:
                res = MonitoredResource(**resource)
        else:
            return None

        return EventHandler.from_resource(res)

    def stop(self):
        self._observer.stop()
        self._observer.join()
        super().stop()

    def main(self):
        self._observer.start()
        self.wait_stop()



from platypush.message.event import Event


class CustomEvent(Event):
    """
    This type can be used to fire custom events upon which the user can implement custom hooks.
    """
    def __init__(self, subtype: str, *args, **kwargs):
        """
        :param subtype: This is the only mandatory attribute for this event type. It should be a string that
            unambiguously identifies a certain type of event (like ``DISHWASHER_STARTED`` or ``SMOKE_DETECTED``).
        :param args: Extra list arguments for the event.
        :param kwargs: Extra key-value arguments for the event.
        """
        super().__init__(*args, subtype=subtype, **kwargs)


# vim:sw=4:ts=4:et:



import base64
import json
import logging
from collections import defaultdict
from threading import RLock, Thread
from typing import Dict, Iterable, Optional, Union

from kafka import KafkaConsumer, KafkaProducer

from platypush.message.event.kafka import KafkaMessageEvent
from platypush.plugins import RunnablePlugin, action


class KafkaPlugin(RunnablePlugin):
    """
    Plugin to send messages to an Apache Kafka instance (https://kafka.apache.org/)
    """

    def __init__(
        self,
        host: Optional[str] = None,
        port: int = 9092,
        listeners: Optional[Iterable[dict]] = None,
        connection_retry_secs: float = 5.0,
        **kwargs,
    ):
        """
        :param host: Default Kafka server name or address. If None (default),
            then it has to be specified when calling the ``send_message`` action.
        :param port: Default Kafka server port (default: 9092).
        :param connection_retry_secs: Seconds to wait before retrying to
            connect to the Kafka server after a connection error (default: 5).
        :param listeners: If specified, the Kafka plugin will listen for
            messages on these topics. Use this parameter if you also want to
            listen on other Kafka brokers other than the primary one. This
            parameter supports a list of maps, where each item supports the
            same arguments passed to the main configuration (host, port, topic,
            password etc.). If host/port are omitted, then the host/port value
            from the plugin configuration will be used. If any of the other
            fields are omitted, then their default value will be used (usually
            null). Example:

                .. code-block:: yaml

                    listeners:
                        # This listener use the default configured host/port
                        - topics:
                              - topic1
                              - topic2
                              - topic3

                        # This will use a custom MQTT broker host
                        - host: sensors
                          port: 19200
                          username: myuser
                          password: secret
                          topics:
                              - topic4
                              - topic5

        """

        super().__init__(**kwargs)

        self.host = host
        self.port = port
        self._conn_retry_secs = connection_retry_secs
        self._listeners = listeners or []

        # `server:port` -> KafkaProducer mapping
        self.producers: Dict[str, KafkaProducer] = {}
        # `server:port` -> KafkaConsumer mapping
        self.consumers: Dict[str, KafkaConsumer] = {}

        # Synchronization locks for the producers/consumers maps,
        # since python-kafka is not thread-safe
        self._producers_locks = defaultdict(RLock)
        self._consumers_locks = defaultdict(RLock)

        # Kafka can be very noisy
        logging.getLogger('kafka').setLevel(logging.ERROR)

    def _get_srv_str(
        self, host: Optional[str] = None, port: Optional[int] = None
    ) -> str:
        if not host:
            host = self.host

        assert host, 'No Kafka server specified'
        if not port:
            port = self.port

        return f'{host}:{port}'

    def _get_producer(
        self, host: Optional[str] = None, port: Optional[int] = None, **kwargs
    ):
        srv_str = self._get_srv_str(host, port)
        with self._producers_locks[srv_str]:
            if srv_str not in self.producers:
                self.producers[srv_str] = KafkaProducer(
                    bootstrap_servers=srv_str, **kwargs
                )
            return self.producers[srv_str]

    def _get_consumer(
        self, host: Optional[str] = None, port: Optional[int] = None, **kwargs
    ):
        srv_str = self._get_srv_str(host, port)
        with self._consumers_locks[srv_str]:
            if srv_str not in self.consumers:
                self.consumers[srv_str] = KafkaConsumer(
                    bootstrap_servers=srv_str, **kwargs
                )
            return self.consumers[srv_str]

    def _on_msg(self, record, host: str, port: int):
        try:
            msg = record.value.decode()
        except UnicodeDecodeError:
            msg = base64.b64encode(record.value).decode()

        try:
            msg = json.loads(msg)
        except (TypeError, ValueError):
            pass

        self._bus.post(
            KafkaMessageEvent(msg=msg, topic=record.topic, host=host, port=port)
        )

    def _consumer_monitor(self, consumer: KafkaConsumer, host: str, port: int):
        while not self.should_stop():
            try:
                for msg in consumer:
                    self._on_msg(msg, host=host, port=port)
                    if self.should_stop():
                        break
            except Exception as e:
                if not self.should_stop():
                    self.logger.exception(e)
                    self.logger.warning(
                        'Kafka connection error to %s:%d, reconnecting in %f seconds',
                        host,
                        port,
                        self._conn_retry_secs,
                    )

                self.wait_stop(self._conn_retry_secs)

    @action
    def publish(self, msg: Union[str, list, dict, tuple, bytes], topic: str, **kwargs):
        """
        :param msg: Message to send.
        :param topic: Topic to send the message to.
        :param kwargs: Additional arguments to pass to the KafkaConsumer,
            including ``host`` and ``port``.
        """
        if isinstance(msg, tuple):
            msg = list(msg)
        if isinstance(msg, (dict, list)):
            msg = json.dumps(msg)
        if not isinstance(msg, bytes):
            msg = str(msg).encode()

        producer = self._get_producer(**kwargs)
        producer.send(topic, msg)
        producer.flush()

    @action
    def send_message(
        self, msg: Union[str, list, dict, tuple, bytes], topic: str, **kwargs
    ):
        """
        Alias for :meth:`.publish`.
        """
        return self.send_message(msg=msg, topic=topic, **kwargs)

    @action
    def subscribe(self, topic: str, **kwargs):
        """
        Subscribe to a topic.

        :param topic: Topic to subscribe to.
        :param kwargs: Additional arguments to pass to the KafkaConsumer,
            including ``host`` and ``port``.
        """
        consumer = self._get_consumer(**kwargs)
        consumer.subscribe([topic])

    @action
    def unsubscribe(self, **kwargs):
        """
        Unsubscribe from all the topics on a consumer.

        :param kwargs: Additional arguments to pass to the KafkaConsumer,
            including ``host`` and ``port``.
        """
        consumer = self._get_consumer(**kwargs)
        consumer.unsubscribe()

    def main(self):
        for listener in self._listeners:
            host = listener.get('host', self.host)
            port = listener.get('port', self.port)
            topics = listener.get('topics')
            if not topics:
                continue

            consumer = self._get_consumer(
                host=host,
                port=port,
                group_id='platypush',
                auto_offset_reset='earliest',
            )

            consumer.subscribe(topics)
            Thread(
                target=self._consumer_monitor,
                args=(consumer,),
                kwargs={'host': host, 'port': port},
                daemon=True,
            ).start()

        self.wait_stop()

    def stop(self):
        super().stop()
        for srv, producer in self.producers.items():
            try:
                producer.flush()
                producer.close()
            except Exception as e:
                self.logger.warning('Error while closing Kafka producer %s: %s', srv, e)

        for srv, consumer in self.consumers.items():
            try:
                consumer.close()
            except Exception as e:
                self.logger.warning('Error while closing Kafka consumer %s: %s', srv, e)


# vim:sw=4:ts=4:et:



import argparse
from typing import Sequence

from platypush.bus.redis import RedisBus
from platypush.utils import get_default_pid_file


def parse_cmdline(args: Sequence[str]) -> argparse.Namespace:
    """
    Parse command-line arguments from a list of strings.
    """
    parser = argparse.ArgumentParser(
        'platypush',
        description='A general-purpose platform for automation. See https://docs.platypush.tech for more info.',
    )

    parser.add_argument(
        '--config',
        '-c',
        dest='config',
        required=False,
        default=None,
        help='Custom location for the configuration file',
    )

    parser.add_argument(
        '--workdir',
        '-w',
        dest='workdir',
        required=False,
        default=None,
        help='Custom working directory to be used for the application',
    )

    parser.add_argument(
        '--main-db',
        '--db',
        dest='db_engine',
        required=False,
        default=None,
        help='Custom database engine to be used for the application '
        '(e.g. sqlite:///:memory: or sqlite:///path/to/db.sqlite). '
        'If missing, it falls back to the value of the `main.db` setting in the configuration file. '
        'If missing, it falls back to sqlite://<workdir>/main.db.',
    )

    parser.add_argument(
        '--cachedir',
        dest='cachedir',
        required=False,
        default=None,
        help='Custom cache directory',
    )

    parser.add_argument(
        '--device-id',
        '-d',
        dest='device_id',
        required=False,
        default=None,
        help='Override the device ID used to identify this instance. If not '
        'passed here, it is inferred from the configuration (device_id field).'
        'If not present there either, it is inferred from the hostname.',
    )

    parser.add_argument(
        '--logsdir',
        '-l',
        dest='logsdir',
        required=False,
        default=None,
        help='Store logs in the specified directory. By default, the '
        '`[logging.]filename` configuration option will be used. If not '
        'set, logging will be sent to stdout and stderr.',
    )

    parser.add_argument(
        '--version',
        dest='version',
        required=False,
        action='store_true',
        help="Print the current version and exit",
    )

    parser.add_argument(
        '--verbose',
        '-v',
        dest='verbose',
        required=False,
        action='store_true',
        help="Enable verbose/debug logging",
    )

    parser.add_argument(
        '--pidfile',
        '-P',
        dest='pidfile',
        required=False,
        default=get_default_pid_file(),
        help="File where platypush will "
        + "store its PID, useful if you're planning to "
        + f"integrate it in a service (default: {get_default_pid_file()})",
    )

    parser.add_argument(
        '--no-capture-stdout',
        dest='no_capture_stdout',
        required=False,
        action='store_true',
        help="Set this flag if you have max stack depth "
        + "exceeded errors so stdout won't be captured by "
        + "the logging system",
    )

    parser.add_argument(
        '--no-capture-stderr',
        dest='no_capture_stderr',
        required=False,
        action='store_true',
        help="Set this flag if you have max stack depth "
        + "exceeded errors so stderr won't be captured by "
        + "the logging system",
    )

    parser.add_argument(
        '--redis-queue',
        dest='redis_queue',
        required=False,
        default=RedisBus.DEFAULT_REDIS_QUEUE,
        help="Name of the Redis queue to be used to internally deliver messages "
        f"(default: {RedisBus.DEFAULT_REDIS_QUEUE})",
    )

    parser.add_argument(
        '--start-redis',
        dest='start_redis',
        required=False,
        action='store_true',
        help="Set this flag if you want to run and manage Redis internally "
        "from the app rather than using an external server. It requires the "
        "redis-server executable to be present in the path",
    )

    parser.add_argument(
        '--redis-host',
        dest='redis_host',
        required=False,
        default=None,
        help="Overrides the host specified in the redis section of the "
        "configuration file",
    )

    parser.add_argument(
        '--redis-port',
        dest='redis_port',
        required=False,
        default=None,
        help="Overrides the port specified in the redis section of the "
        "configuration file",
    )

    parser.add_argument(
        '--ctrl-sock',
        dest='ctrl_sock',
        required=False,
        default=None,
        help="If set, it identifies a path to a UNIX domain socket that "
        "the application can use to send control messages (e.g. STOP and "
        "RESTART) to its parent.",
    )

    opts, _ = parser.parse_known_args(args)
    return opts



from abc import ABC, abstractmethod

import aioxmpp

from .._mixins import XmppAsyncMixin, XmppConfigMixin, XmppEventStateMixin


# pylint: disable=too-few-public-methods
class XmppBaseHandler(XmppAsyncMixin, XmppConfigMixin, XmppEventStateMixin, ABC):
    """
    Base class for XMPP handlers.
    """

    _client: aioxmpp.Client

    @abstractmethod
    def __init__(self, *args, **kwargs):
        """
        To be implemented by the subclasses.
        """
        super().__init__(*args, **kwargs)



from typing import Iterable, List

from sqlalchemy import (
    Boolean,
    Column,
    ForeignKey,
    Integer,
    JSON,
    String,
)

from platypush.common.db import is_defined

from ..devices import Device
from ._service import BluetoothService


if not is_defined('bluetooth_device'):

    class BluetoothDevice(Device):
        """
        Entity that represents a Bluetooth device.
        """

        __tablename__ = 'bluetooth_device'

        id = Column(
            Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True
        )

        address = Column(String, nullable=False)
        """ Device address. """

        connected = Column(Boolean, default=False)
        """ Whether the device is connected. """

        supports_ble = Column(Boolean, default=False)
        """
        Whether the device supports the Bluetooth Low-Energy specification.
        """

        supports_legacy = Column(Boolean, default=False)
        """
        Whether the device supports the legacy (non-BLE) specification.
        """

        rssi = Column(Integer, default=None)
        """ Received Signal Strength Indicator. """

        tx_power = Column(Integer, default=None)
        """ Reported transmission power. """

        _major_service_classes = Column("major_service_classes", JSON, default=None)
        """ The reported major service classes, as a list of strings. """

        _major_device_class = Column("major_device_class", String, default=None)
        """ The reported major device class. """

        _minor_device_classes = Column("minor_device_classes", JSON, default=None)
        """ The reported minor device classes, as a list of strings. """

        manufacturer = Column(String, default=None)
        """ Device manufacturer, as a string. """

        model = Column(String, default=None)
        """ Device model, as a string. """

        model_id = Column(String, default=None)
        """ Device model ID. """

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }

        @property
        def major_device_class(self):
            from platypush.plugins.bluetooth.model import MajorDeviceClass

            ret = MajorDeviceClass.UNKNOWN
            if self._major_device_class:
                matches = [
                    cls
                    for cls in MajorDeviceClass
                    if cls.value.name == self._major_device_class
                ]

                if matches:
                    ret = matches[0]

            return ret

        @major_device_class.setter
        def major_device_class(self, value):
            self._major_device_class = value.value.name

        @property
        def minor_device_classes(self) -> list:
            from platypush.plugins.bluetooth.model import MinorDeviceClass

            ret = []
            for dev_cls in self._minor_device_classes or []:
                matches = [cls for cls in MinorDeviceClass if cls.value.name == dev_cls]

                if matches:
                    ret.append(matches[0])

            return ret

        @minor_device_classes.setter
        def minor_device_classes(self, value: Iterable):
            self._minor_device_classes = [cls.value.name for cls in (value or [])]

        @property
        def major_service_classes(self) -> list:
            from platypush.plugins.bluetooth.model import MajorServiceClass

            ret = []
            for dev_cls in self._major_service_classes or []:
                matches = [
                    cls for cls in MajorServiceClass if cls.value.name == dev_cls
                ]

                if matches:
                    ret.append(matches[0])

            return ret

        @major_service_classes.setter
        def major_service_classes(self, value: Iterable):
            self._major_service_classes = [cls.value.name for cls in (value or [])]

        @property
        def services(self) -> List[BluetoothService]:
            """
            :return: List of
            :class:`platypush.plugins.bluetooth.model.BluetoothService` mapping
            all the services exposed by the device.
            """
            return [
                child for child in self.children if isinstance(child, BluetoothService)
            ]

        @property
        def known_services(self) -> dict:
            """
            Known services exposed by the device, indexed by
            :class:`platypush.plugins.bluetooth.model.ServiceClass` enum value.
            """
            from platypush.plugins.bluetooth.model import ServiceClass

            return {
                child.service_class: child
                for child in self.children
                if isinstance(child, BluetoothService)
                and child.service_class != ServiceClass.UNKNOWN
            }

        def to_dict(self):
            """
            Overwrites ``to_dict`` to transform private column names into their
            public representation.
            """
            return {k.lstrip('_'): v for k, v in super().to_dict().items()}



import json
import os
from typing import Optional

from platypush.config import Config
from platypush.plugins import RunnablePlugin, action
from platypush.plugins.assistant import AssistantPlugin


class AssistantGooglePlugin(AssistantPlugin, RunnablePlugin):
    """
    Google Assistant plugin.

    This plugin allows you to run the Google Assistant _directly_ on your
    device. It requires you to have an audio microphone and a speaker connected
    to the device.

    If you have multiple sound devices, you can specify which one(s) to use for
    input and output through a ``~/.asoundrc`` configuration file like this:

        .. code-block:: text

            pcm.!default {
              type asym
               playback.pcm {
                 type plug
                 slave.pcm "hw:0,0"
               }
               capture.pcm {
                 type plug
                 slave.pcm "hw:1,0"
               }
            }

    You can use ``aplay -l`` and ``arecord -l`` to respectively list the
    detected audio output and input devices with their indices.

    If you are using PulseAudio instead of bare ALSA, then you can:

        1. Use the ``pavucontrol`` (GUI) tool to select the audio input and
           output devices and volumes for the assistant.
        2. Use a program like ``pamix`` (ncurses) or ``pamixer`` (CLI).
        3. Run the ``pactl list sources`` and ``pactl list sinks`` commands to
           respectively list the detected audio input and output devices. Take
           note of their name, and specify which ones the assistant should use
           by starting the application with the right ``PULSE_SOURCE`` and
           ``PULSE_SINK`` environment variables.

    .. warning:: The Google Assistant library used by this backend has
        been deprecated by Google:
        https://developers.google.com/assistant/sdk/reference/library/python/.
        This integration still works on all of my devices, but its future
        functionality is not guaranteed - Google may decide to turn off the
        API, the library may no longer be built against new architectures and
        it's unlikely to be updated.

    .. note:: Since the Google Assistant library hasn't been updated in several
        years, some of its dependencies are quite old and may break more recent
        Python installations. Please refer to the comments in the `manifest
        file
        <https://git.platypush.tech/platypush/platypush/src/branch/master/platypush/plugins/assistant/google/manifest.json>`_.
        for more information on how to install the required dependencies, if
        the automated ways fail.
    """

    _entity_name = 'Google Assistant'

    _default_credentials_files = (
        os.path.join(Config.get_workdir(), 'credentials', 'google', 'assistant.json'),
        os.path.join(
            os.path.expanduser('~/.config'), 'google-oauthlib-tool', 'credentials.json'
        ),
    )

    def __init__(
        self,
        credentials_file: Optional[str] = None,
        device_model_id: str = 'Platypush',
        **kwargs,
    ):
        """
        :param credentials_file: Path to the Google OAuth credentials file.
            See
            https://developers.google.com/assistant/sdk/guides/library/python/embed/install-sample#generate_credentials
            for instructions to get your own credentials file.
            By default, it will search for the credentials file under:

                * ``~/.config/google-oauthlib-tool/credentials.json``: default
                  location supported by the Google Assistant library

                * ``<PLATYPUSH_WORKDIR>/credentials/google/assistant.json``:
                  recommended location, under the Platypush working directory.

        :param device_model_id: The device model ID that identifies the device
            where the assistant is running (default: Platypush). It can be a
            custom string.
        """
        super().__init__(**kwargs)
        self._credentials_file = credentials_file
        self.device_model_id = device_model_id
        self.credentials = None
        self._assistant = None
        self.logger.info('Initialized Google Assistant plugin')

    @property
    def credentials_file(self) -> str:
        if self._credentials_file:
            return os.path.abspath(os.path.expanduser(self._credentials_file))

        f = None
        for default_file in self._default_credentials_files:
            f = default_file
            if os.path.isfile(default_file):
                break

        assert f, 'No credentials_file provided and no default file found'
        return f

    @property
    def assistant(self):
        if not self._assistant:
            self.logger.warning('The assistant is not running')
            return

        return self._assistant

    def _process_event(self, event):
        from google.assistant.library.event import EventType, AlertType

        self.logger.info('Received assistant event: %s', event)

        if event.type == EventType.ON_CONVERSATION_TURN_STARTED:
            self._on_conversation_start()
        elif event.type == EventType.ON_CONVERSATION_TURN_FINISHED:
            if not event.args.get('with_follow_on_turn'):
                self._on_conversation_end()
        elif event.type == EventType.ON_CONVERSATION_TURN_TIMEOUT:
            self._on_conversation_timeout()
        elif event.type == EventType.ON_NO_RESPONSE:
            self._on_no_response()
        elif (
            hasattr(EventType, 'ON_RENDER_RESPONSE')
            and event.type == EventType.ON_RENDER_RESPONSE
        ):
            self._on_response_render_start(event.args.get('text'))
        elif (
            hasattr(EventType, 'ON_RESPONDING_STARTED')
            and event.type == EventType.ON_RESPONDING_STARTED
            and event.args.get('is_error_response') is True
        ):
            self.logger.warning('Assistant response error: %s', json.dumps(event.args))
        elif event.type == EventType.ON_RECOGNIZING_SPEECH_FINISHED:
            self._on_speech_recognized(event.args.get('text'))
        elif event.type == EventType.ON_ALERT_STARTED:
            if event.args.get('alert_type') == AlertType.ALARM:
                self._on_alarm_start()
            elif event.args.get('alert_type') == AlertType.TIMER:
                self._on_timer_start()
            else:
                self._on_alert_start()
        elif event.type == EventType.ON_ALERT_FINISHED:
            if event.args.get('alert_type') == AlertType.ALARM:
                self._on_alarm_end()
            elif event.args.get('alert_type') == AlertType.TIMER:
                self._on_timer_end()
            else:
                self._on_alert_end()
        elif event.type == EventType.ON_ASSISTANT_ERROR:
            if event.args.get('is_fatal'):
                raise RuntimeError(f'Fatal assistant error: {json.dumps(event.args)}')

            self.logger.warning('Assistant error: %s', json.dumps(event.args))
        elif event.type == EventType.ON_MUTED_CHANGED:
            self._on_mute_changed(event.args.get('is_muted', False))

    @action
    def start_conversation(self, *_, **__):
        """
        Programmatically start a conversation with the assistant
        """
        if self.assistant:
            self.assistant.start_conversation()

    def _stop_conversation(self, *_, **__):
        """
        Programmatically stop a running conversation with the assistant
        """
        if self.assistant:
            self.assistant.stop_conversation()

    @action
    def mute(self, *_, **__):
        """
        Mute the microphone. Alias for :meth:`.set_mic_mute` with
        ``muted=True``.
        """
        self.set_mic_mute(muted=True)

    @action
    def unmute(self, *_, **__):
        """
        Unmute the microphone. Alias for :meth:`.set_mic_mute` with
        ``muted=False``.
        """
        self.set_mic_mute(muted=False)

    @action
    def set_mic_mute(self, muted: bool):
        """
        Programmatically mute/unmute the microphone.

        :param muted: Set to True or False.
        """
        if self.assistant:
            self.assistant.set_mic_mute(muted)

            if muted:
                self._on_mute()
            else:
                self._on_unmute()

    @action
    def toggle_mute(self, *_, **__):
        """
        Toggle the mic mute state.
        """
        self.set_mic_mute(muted=not self._is_muted)

    @action
    def toggle_mic_mute(self):
        """
        Deprecated alias for :meth:`.toggle_mute`.
        """
        return self.toggle_mute()

    @action
    def send_text_query(self, *_, query: str, **__):
        """
        Send a text query to the assistant.

        This is equivalent to saying something to the assistant.

        :param query: Query to be sent.
        """
        if self.assistant:
            self.assistant.send_text_query(query)

    def main(self):
        import google.oauth2.credentials
        from google.assistant.library import Assistant

        last_sleep = 0

        while not self.should_stop():
            try:
                with open(self.credentials_file, 'r') as f:
                    self.credentials = google.oauth2.credentials.Credentials(
                        token=None, **json.load(f)
                    )
            except Exception as e:
                self.logger.error(
                    'Error while loading Google Assistant credentials: %s', e
                )
                self.logger.info(
                    'Please follow the instructions at '
                    'https://developers.google.com/assistant/sdk/guides/library/python/embed/install-sample'
                    '#generate_credentials to get your own credentials file'
                )
                self.logger.exception(e)
                break

            try:
                with Assistant(
                    self.credentials, self.device_model_id
                ) as self._assistant:
                    self.publish_entities([self])
                    for event in self._assistant.start():
                        last_sleep = 0

                        if not self.is_detecting():
                            self.logger.info(
                                'Assistant event received but detection is currently paused'
                            )
                            continue

                        self._process_event(event)
            except Exception as e:
                self.logger.exception(e)
                sleep_secs = min(60, max(5, last_sleep * 2))
                self.logger.warning(
                    'Restarting the assistant in %d seconds after an unrecoverable error',
                    sleep_secs,
                )

                self.wait_stop(sleep_secs)
                last_sleep = sleep_secs
                continue

    def stop(self):
        try:
            self.stop_conversation()
        except RuntimeError:
            pass

        if self._assistant:
            del self._assistant
            self._assistant = None

        super().stop()


# vim:sw=4:ts=4:et:



import base64
from functools import wraps
from typing import Optional

from flask import request, redirect, jsonify
from flask.wrappers import Response

from platypush.config import Config
from platypush.user import UserManager

from ..logger import logger
from .status import AuthStatus

user_manager = UserManager()


def get_arg(req, name: str) -> Optional[str]:
    # The Flask way
    if hasattr(req, 'args'):
        return req.args.get(name)

    # The Tornado way
    if hasattr(req, 'arguments'):
        arg = req.arguments.get(name)
        if arg:
            return arg[0].decode()

    return None


def get_cookie(req, name: str) -> Optional[str]:
    cookie = req.cookies.get(name)
    if not cookie:
        return None

    # The Flask way
    if isinstance(cookie, str):
        return cookie

    # The Tornado way
    return cookie.value


def authenticate_token(req):
    token = Config.get('token')
    user_token = None

    if 'X-Token' in req.headers:
        user_token = req.headers['X-Token']
    elif 'Authorization' in req.headers and req.headers['Authorization'].startswith(
        'Bearer '
    ):
        user_token = req.headers['Authorization'][7:]
    else:
        user_token = get_arg(req, 'token')

    if not user_token:
        return False

    try:
        user_manager.validate_jwt_token(user_token)
        return True
    except Exception as e:
        logger().debug(str(e))
        return bool(token and user_token == token)


def authenticate_user_pass(req):
    # Flask populates request.authorization
    if hasattr(req, 'authorization'):
        if not req.authorization:
            return False

        username = req.authorization.username
        password = req.authorization.password

    # Otherwise, check the Authorization header
    elif 'Authorization' in req.headers and req.headers['Authorization'].startswith(
        'Basic '
    ):
        auth = req.headers['Authorization'][6:]
        try:
            auth = base64.b64decode(auth)
        except ValueError:
            pass

        username, password = auth.decode().split(':', maxsplit=1)
    else:
        return False

    return user_manager.authenticate_user(username, password)


def authenticate_session(req):
    user = None

    # Check the X-Session-Token header
    user_session_token = req.headers.get('X-Session-Token')

    # Check the `session_token` query/body parameter
    if not user_session_token:
        user_session_token = get_arg(req, 'session_token')

    # Check the `session_token` cookie
    if not user_session_token:
        user_session_token = get_cookie(req, 'session_token')

    if user_session_token:
        user, _ = user_manager.authenticate_user_session(user_session_token)

    return user is not None


def authenticate(
    redirect_page='',
    skip_auth_methods=None,
    json=False,
):
    """
    Authentication decorator for Flask routes.
    """

    def decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            auth_status = get_auth_status(
                request,
                skip_auth_methods=skip_auth_methods,
            )

            if auth_status == AuthStatus.OK:
                return f(*args, **kwargs)

            if json:
                return jsonify(auth_status.to_dict()), auth_status.value.code

            if auth_status == AuthStatus.NO_USERS:
                return redirect(
                    f'/register?redirect={redirect_page or request.url}', 307
                )

            if auth_status == AuthStatus.UNAUTHORIZED:
                return redirect(f'/login?redirect={redirect_page or request.url}', 307)

            return Response(
                'Authentication required',
                401,
                {'WWW-Authenticate': 'Basic realm="Login required"'},
            )

        return wrapper

    return decorator


# pylint: disable=too-many-return-statements
def get_auth_status(req, skip_auth_methods=None) -> AuthStatus:
    """
    Check against the available authentication methods (except those listed in
    ``skip_auth_methods``) if the user is properly authenticated.
    """

    n_users = user_manager.get_user_count()
    skip_methods = skip_auth_methods or []

    # User/pass HTTP authentication
    http_auth_ok = True
    if n_users > 0 and 'http' not in skip_methods:
        http_auth_ok = authenticate_user_pass(req)
        if http_auth_ok:
            return AuthStatus.OK

    # Token-based authentication
    token_auth_ok = True
    if 'token' not in skip_methods:
        token_auth_ok = authenticate_token(req)
        if token_auth_ok:
            return AuthStatus.OK

    # Session token based authentication
    session_auth_ok = True
    if n_users > 0 and 'session' not in skip_methods:
        return AuthStatus.OK if authenticate_session(req) else AuthStatus.UNAUTHORIZED

    # At least a user should be created before accessing an authenticated resource
    if n_users == 0 and 'session' not in skip_methods:
        return AuthStatus.NO_USERS

    if (  # pylint: disable=too-many-boolean-expressions
        ('http' not in skip_methods and http_auth_ok)
        or ('token' not in skip_methods and token_auth_ok)
        or ('session' not in skip_methods and session_auth_ok)
    ):
        return AuthStatus.OK

    return AuthStatus.UNAUTHORIZED



from logging import getLogger
from socket import socket
from typing import Optional

from platypush.commands import Command


# pylint: disable=too-few-public-methods
class CommandReader:
    """
    Reads command objects from file-like I/O objects.
    """

    _max_bufsize = 8192
    """Maximum size of a command that can be queued in the stream."""

    _bufsize = 1024
    """Size of the buffer used to read commands from the socket."""

    def __init__(self):
        self.logger = getLogger(__name__)
        self._buf = bytes()

    def _parse_command(self, data: bytes) -> Optional[Command]:
        """
        Parses a command from the received data.

        :param data: Data received from the socket
        :return: The parsed command
        """
        try:
            return Command.parse(data)
        except ValueError as e:
            self.logger.warning('Error while parsing command: %s', e)
            return None

    def read(self, sock: socket) -> Optional[Command]:
        """
        Parses the next command from the file-like I/O object.

        :param fp: The file-like I/O object to read from.
        :return: The parsed command.
        """
        try:
            data = sock.recv(self._bufsize)
        except OSError as e:
            self.logger.warning(
                'Error while reading from socket %s: %s', sock.getsockname(), e
            )
            return None

        for ch in data:
            if bytes([ch]) == Command.END_OF_COMMAND:
                cmd = self._parse_command(self._buf)
                self._buf = bytes()

                if cmd:
                    return cmd
            elif len(self._buf) >= self._max_bufsize:
                self.logger.warning(
                    'The received command is too long: length=%d', len(self._buf)
                )

                self._buf = bytes()
                break
            else:
                self._buf += bytes([ch])

        return None



import datetime
import json
import logging
import multiprocessing
import os
import subprocess
from concurrent.futures import ProcessPoolExecutor

from dateutil.parser import isoparse
import shutil

logger = logging.getLogger(__name__)


def get_file_metadata(path: str):
    """
    Retrieves the metadata of a media file using ffprobe.
    """
    logger.info('Retrieving metadata for %s', path)

    with subprocess.Popen(
        [
            'ffprobe',
            '-v',
            'quiet',
            '-print_format',
            'json',
            '-show_format',
            '-show_streams',
            path,
        ],
        stdout=subprocess.PIPE,
    ) as ffprobe:
        ret = json.loads(ffprobe.communicate()[0])

    video_stream = next(
        iter(
            [
                stream
                for stream in ret.get('streams', [])
                if stream.get('codec_type') == 'video'
            ]
        ),
        {},
    )

    creation_time = ret.get('format', {}).get('tags', {}).get('creation_time')
    if creation_time:
        try:
            creation_time = isoparse(creation_time)
        except ValueError:
            creation_time = None

    if not creation_time:
        creation_time = datetime.datetime.fromtimestamp(os.path.getctime(path))

    return {
        'duration': ret.get('format', {}).get('duration'),
        'width': video_stream.get('width'),
        'height': video_stream.get('height'),
        'created_at': creation_time,
    }


def get_metadata(*paths: str):
    """
    Retrieves the metadata of media files using ffprobe.
    """
    logger.info('Retrieving metadata for %d media files', len(paths))
    try:
        assert shutil.which(
            'ffprobe'
        ), 'ffprobe not found in PATH. Install ffmpeg to retrieve local media metadata.'

        # Run ffprobe in parallel
        with ProcessPoolExecutor(
            max_workers=multiprocessing.cpu_count() * 2
        ) as executor:
            futures = [executor.submit(get_file_metadata, path) for path in paths]
            results = [future.result() for future in futures]

        logger.info('Retrieved metadata for %d media files', len(results))
        return results
    except Exception as e:
        logger.error('Failed to retrieve media metadata: %s: %s', type(e).__name__, e)
        return [{} for _ in paths]



from dataclasses import dataclass, field
from typing import Collection, Union
from uuid import UUID

RawServiceClass = Union[UUID, int]
"""
Raw type for service classes received by pybluez.
Can be either a 16-bit integer or a UUID.
"""


@dataclass
class DevicesBlacklist:
    """
    A data class representing the rules for blacklisting devices.
    """

    addresses: Collection[str] = field(default_factory=set[str])
    """ MAC addresses to ignore. """
    names: Collection[str] = field(default_factory=set[str])
    """ Device names to ignore. """
    manufacturers: Collection[str] = field(default_factory=set[str])
    """ Manufacturers strings to ignore. """

    def __post_init__(self):
        """
        Normalize the blacklist rules.
        """
        # Normalize case-sensitivity for addresses
        self.addresses = set(map(str.lower, self.addresses))
        # Make sure that all the collections are sets
        self.names = set(self.names)
        self.manufacturers = set(self.manufacturers)

    def matches(self, device) -> bool:
        """
        :type device: :class:`platypush.entities.bluetooth.BluetoothDevice`.
        """
        return (
            device.address.lower() in self.addresses
            or device.name in self.names
            or device.manufacturer in self.manufacturers
        )



from dataclasses import dataclass, field
from logging import getLogger
from threading import Event, Thread, current_thread
from typing import Optional

import bluetooth

from platypush.entities.bluetooth import BluetoothService

from ._types import ConnectionId

_logger = getLogger(__name__)


@dataclass
class BluetoothConnection:
    """
    Models a connection to a Bluetooth device.
    """

    address: str
    service: BluetoothService
    socket: bluetooth.BluetoothSocket = field(
        init=False, default_factory=bluetooth.BluetoothSocket
    )
    thread: Thread = field(default_factory=current_thread)
    stop_event: Event = field(default_factory=Event)

    def __post_init__(self):
        """
        Initialize the Bluetooth socket with the given protocol.
        """
        self.socket = bluetooth.BluetoothSocket(self.service.protocol.value)

    @property
    def key(self) -> ConnectionId:
        return self.address, self.service.port

    def close(self, timeout: Optional[float] = None):
        _logger.info('Closing connection to %s', self.address)

        # Set the stop event
        self.stop_event.set()

        # Close the socket
        if self.socket:
            try:
                self.socket.close()
            except Exception as e:
                _logger.warning(
                    'Failed to close Bluetooth socket on %s: %s',
                    self.address,
                    e,
                )

        # Avoid deadlocking by waiting for our own thread to terminate
        if current_thread() is self.thread:
            return

        # Wait for the connection thread to terminate
        if self.thread and self.thread.is_alive():
            self.thread.join(timeout=timeout)
        if self.thread and self.thread.is_alive():
            _logger.warning('Connection to %s still alive after closing', self.address)



import json
import os
import pathlib
from typing import List, Dict, Optional

from platypush.plugins import Plugin, action


class FilePlugin(Plugin):
    """
    A plugin for general-purpose file methods
    """

    @classmethod
    def _get_path(cls, filename):
        return os.path.abspath(os.path.expanduser(filename))

    @classmethod
    def _to_string(cls, content):
        try:
            return json.dumps(content)
        except (ValueError, TypeError):
            return str(content)

    @action
    def read(self, file: str):
        """
        Read and return the content of a (text) file.

        Note that this method should only be used for small text files, as it
        reads the entire file in memory.

        If you need to read large/binary files, consider using the
        ``GET /file?path=<path>`` HTTP API endpoint instead.

        :param file: Path of the file.
        """
        with open(self._get_path(file), 'r') as f:
            return f.read()

    @action
    def write(self, file: str, content: str):
        """
        Writes content to a specified (text) file. Previous content will be truncated.

        :param file: Path of the file.
        :param content: Content to write.
        """

        if not isinstance(content, str):
            content = self._to_string(content)

        with open(self._get_path(file), 'w') as f:
            f.write(content)

    @action
    def append(self, file: str, content):
        """
        Append content to a specified (text) file.

        :param file: Path of the file.
        :param content: Content to write.
        """

        if not isinstance(content, str):
            content = self._to_string(content)

        with open(self._get_path(file), 'a') as f:
            f.write(content)

    @action
    def getsize(self, file):
        """
        Get the size of the specified file in bytes.

        :param file: File path.
        """
        return os.path.getsize(self._get_path(file))

    @action
    def mkdir(self, directory: str, exist_ok=True, parents=True, mode=0o755):
        """
        Create a directory.

        :param directory: Directory name/path.
        :param exist_ok: If set and the directory already exist the method will
            not return an error (default: True).
        :param parents: If set and any of the parent directories in the path don't
            exist they will be created (analogous to mkdir -p) (default: True).
        :param mode: Access mode (default: 0755).
        """
        pathlib.Path(self._get_path(directory)).mkdir(
            parents=parents, exist_ok=exist_ok, mode=mode
        )

    @action
    def rmdir(self, directory: str):
        """
        Remove a directory. The directory must be empty.

        :param directory: Directory name/path.
        """
        pathlib.Path(self._get_path(directory)).rmdir()

    @action
    def touch(self, file: str, mode=0o644):
        """
        Create/touch a file.

        :param file: File name/path.
        :param mode: File permissions (default: 0644).
        """
        pathlib.Path(self._get_path(file)).touch(mode=mode)

    @action
    def chmod(self, file: str, mode):
        """
        Change the mode/permissions of a file.

        :param file: File name/path.
        :param mode: New file permissions.
        """
        pathlib.Path(self._get_path(file)).chmod(mode=mode)

    @action
    def home(self) -> str:
        """
        Returns the current user's home directory.
        """
        return str(pathlib.Path.home())

    @action
    def rename(self, file: str, name: str):
        """
        Rename/move a file.

        :param file: File to rename.
        :param name: New file name.
        """
        pathlib.Path(self._get_path(file)).rename(self._get_path(name))

    @action
    def link(self, file: str, target: str, symbolic=True):
        """
        Create a link to a file.

        :param file: File to symlink.
        :param target: Symlink path.
        :param symbolic: If True, then the target link will be a symbolic link. Otherwise,
            it will be a hard link (default: symbolic).
        """
        path = pathlib.Path(self._get_path(file))
        target = self._get_path(target)

        if symbolic:
            path.symlink_to(target)
        else:
            path.hardlink_to(target)

    @action
    def unlink(self, file: str):
        """
        Remove a file or symbolic link.

        :param file: File/link to remove.
        """
        pathlib.Path(self._get_path(file)).unlink()

    @action
    def list(self, path: Optional[str] = None) -> List[Dict[str, str]]:
        """
        List a file or all the files in a directory.

        :param path: File or directory (default: root directory).
        :return: List of files in the specified path, or absolute path of the specified path if ``path`` is a file and
            it exists. Each item will contain the fields ``type`` (``file`` or ``directory``) and ``path``.
        """
        path = self._get_path(path or '/')
        assert path and os.path.exists(path), f'No such file or directory: {path}'

        if not os.path.isdir(path):
            return [
                {
                    'type': 'file',
                    'path': path,
                    'name': os.path.basename(path),
                }
            ]

        return sorted(
            [
                {
                    'type': (
                        'directory' if os.path.isdir(os.path.join(path, f)) else 'file'
                    ),
                    'path': os.path.join(path, f),
                    'name': os.path.basename(f),
                }
                for f in sorted(os.listdir(path))
            ],
            key=lambda f: (f.get('type'), f.get('name')),
        )


# vim:sw=4:ts=4:et:



from ._model import Cpu, CpuFrequency, CpuInfo, CpuStats, CpuTimes
from ._schemas import CpuFrequencySchema, CpuInfoSchema, CpuStatsSchema, CpuTimesSchema


__all__ = [
    "Cpu",
    "CpuFrequency",
    "CpuFrequencySchema",
    "CpuInfo",
    "CpuInfoSchema",
    "CpuStats",
    "CpuStatsSchema",
    "CpuTimes",
    "CpuTimesSchema",
]



from collections import namedtuple
from enum import Enum


StatusValue = namedtuple('StatusValue', ['code', 'message'])


class AuthStatus(Enum):
    """
    Models the status of the authentication.
    """

    OK = StatusValue(200, 'OK')
    UNAUTHORIZED = StatusValue(401, 'Unauthorized')
    NO_USERS = StatusValue(412, 'Please create a user first')

    def to_dict(self):
        return {
            'code': self.value[0],
            'message': self.value[1],
        }



import logging

from marshmallow import Schema, fields, pre_dump, post_dump

from platypush.context import get_plugin

from . import MediaArtistSchema, MediaCollectionSchema, MediaVideoSchema

logger = logging.getLogger(__name__)


class JellyfinSchema(Schema):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if 'id' in self.fields:
            self.fields['id'].attribute = 'Id'
        if 'name' in self.fields:
            self.fields['name'].attribute = 'Name'
        elif 'title' in self.fields:
            self.fields['title'].attribute = 'Name'

    @post_dump
    def gen_img_url(self, data: dict, **_) -> dict:
        if 'image' in self.fields:
            plugin = get_plugin('media.jellyfin')
            assert plugin, 'The media.jellyfin plugin is not configured'
            data['image'] = (
                plugin.server + f'/Items/{data["id"]}'  # type: ignore
                '/Images/Primary?fillHeight=333&fillWidth=222&quality=96'
            )

        return data

    @pre_dump
    def _gen_video_url(self, data, **_):
        if data.get('MediaType') != 'Video':
            return data

        video_format = None
        containers_priority = ['mp4', 'mkv', 'm4a', 'mov', 'avi']
        available_containers = data.get('Container', '').split(',')
        for container in containers_priority:
            if container in available_containers:
                video_format = container
                break

        if not video_format:
            if not available_containers:
                logger.warning(
                    'The media ID %s has no available video containers', data["Id"]
                )

                return data

            video_format = available_containers[0]

        plugin = get_plugin('media.jellyfin')
        assert plugin, 'The media.jellyfin plugin is not configured'
        data['url'] = (
            f'{plugin.server}/Videos/{data["Id"]}'
            f'/stream.{video_format}'
            f'?Static=true&api_key={plugin._api_key}'
        )

        return data


class JellyfinArtistSchema(JellyfinSchema, MediaArtistSchema):
    pass


class JellyfinCollectionSchema(JellyfinSchema, MediaCollectionSchema):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['type'].attribute = 'CollectionType'


class JellyfinVideoSchema(JellyfinSchema, MediaVideoSchema):
    duration = fields.Number(attribute='RunTimeTicks')
    community_rating = fields.Number(attribute='CommunityRating')
    critic_rating = fields.Number(attribute='CriticRating')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['year'].attribute = 'ProductionYear'
        self.fields['has_subtitles'].attribute = 'HasSubtitles'

    @post_dump
    def _normalize_community_rating(self, data: dict, **_) -> dict:
        if data.get('community_rating'):
            data['community_rating'] *= 10
        return data

    @post_dump
    def _normalize_duration(self, data: dict, **_) -> dict:
        if data.get('duration'):
            data['duration'] //= 1e7
        return data


class JellyfinMovieSchema(JellyfinVideoSchema):
    pass


class JellyfinEpisodeSchema(JellyfinVideoSchema):
    @pre_dump
    def _normalize_episode_name(self, data: dict, **_) -> dict:
        prefix = ''
        series_name = data.get('SeriesName')
        if series_name:
            prefix = series_name

        episode_index = data.get('IndexNumber')
        if episode_index:
            season_index = data.get('SeasonIndex', 1)
            episode_index = 's{:02d}e{:02d}'.format(season_index, episode_index)

        if episode_index:
            prefix += f'{" " if prefix else ""}[{episode_index}] '

        data['Name'] = prefix + data.get('Name', '')
        return data



from abc import ABC, abstractmethod
from typing import Optional

from .._config import XmppConfig
from ._base import XmppBaseMixin


# pylint: disable=too-few-public-methods
class XmppConfigMixin(XmppBaseMixin, ABC):
    """
    A simple mixin that encapsulates an XMPP configuration object.
    """

    @abstractmethod
    def __init__(self, *args, config: Optional[XmppConfig] = None, **kwargs):
        self._config = config or XmppConfig()
        super().__init__(*args, **kwargs)



import importlib
import inspect
import os
from typing import List, Type

import pkgutil

from ._base import XmppBaseHandler


def discover_handlers() -> List[Type[XmppBaseHandler]]:
    """
    Discover the handler classes defined in this module.
    """

    base_pkg = '.'.join(__name__.split('.')[:-1])
    base_dir = os.path.dirname(__file__)
    return [
        obj
        for _, mod_name, _ in pkgutil.walk_packages([base_dir], prefix=base_pkg + '.')
        for _, obj in inspect.getmembers(importlib.import_module(mod_name))
        if inspect.isclass(obj)
        and not inspect.isabstract(obj)
        and issubclass(obj, XmppBaseHandler)
    ]



import os

from typing import Optional, Dict, Any, List

from platypush.plugins import Plugin, action
from platypush.schemas.cups import JobAddedSchema, PrinterSchema


class CupsPlugin(Plugin):
    """
    A plugin to interact with local and remote printers over a CUPS server.
    """

    def __init__(
        self,
        host: str = 'localhost',
        port: int = 631,
        printer: Optional[str] = None,
        **kwargs
    ):
        """
        :param host: CUPS host IP/name (default: localhost).
        :param port: CUPS port (default: 631).
        :param printer: Default printer name that should be used.
        """
        super().__init__(**kwargs)
        self.host = host
        self.port = port
        self.printer = printer

    def _get_connection(self, host: Optional[str] = None, port: Optional[int] = None):
        import cups

        connection = cups.Connection(host=host or self.host, port=port or self.port)
        return connection

    def _get_printer(self, printer: Optional[str] = None):
        printer = printer or self.printer
        assert printer, 'No printer specified nor default printer configured'
        return printer

    @action
    def get_printers(
        self, host: Optional[str] = None, port: Optional[int] = None
    ) -> List[dict]:
        """
        Get the list of printers registered on a CUPS server.

        :param host: CUPS server host IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        :return: .. schema:: cups.PrinterSchema(many=True)
        """
        conn = self._get_connection(host, port)
        return PrinterSchema().dump(conn.getPrinters())

    @action
    def print_test_page(
        self,
        printer: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ) -> dict:
        """
        Print the CUPS test page.

        :param printer: Printer name (default: default configured ``printer``).
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        :return: .. schema:: cups.JobAddedSchema
        """
        conn = self._get_connection(host, port)
        printer = self._get_printer(printer)
        job_id = conn.printTestPage(printer)
        return dict(JobAddedSchema().dump({'printer': printer, 'job_id': job_id}))

    @action
    def print_file(
        self,
        filename: str,
        printer: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        title: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> dict:
        """
        Print a file.

        :param filename: Path to the file to print.
        :param printer: Printer name (default: default configured ``printer``).
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        :param title: Print title.
        :param options: Extra CUPS name->value options.
        :return: .. schema:: cups.JobAddedSchema
        """
        filename = os.path.abspath(os.path.expanduser(filename))
        conn = self._get_connection(host, port)
        printer = self._get_printer(printer)
        job_id = conn.printFile(
            printer, filename=filename, title=title or '', options=options or {}
        )
        return dict(JobAddedSchema().dump({'printer': printer, 'job_id': job_id}))

    @action
    def print_files(
        self,
        filenames: List[str],
        printer: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
        title: Optional[str] = None,
        options: Optional[Dict[str, Any]] = None,
    ) -> dict:
        """
        Print a list of files.

        :param filenames: Paths to the files to print.
        :param printer: Printer name (default: default configured ``printer``).
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        :param title: Print title.
        :param options: Extra CUPS name->value options.
        :return: .. schema:: cups.JobAddedSchema
        """
        filenames = [os.path.abspath(os.path.expanduser(f)) for f in filenames]
        conn = self._get_connection(host, port)
        printer = self._get_printer(printer)
        job_id = conn.printFiles(
            printer, filenames=filenames, title=title or '', options=options or {}
        )
        return dict(JobAddedSchema().dump({'printer': printer, 'job_id': job_id}))

    @action
    def add_printer(
        self,
        name: str,
        ppd_file: str,
        info: str,
        location: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Add a printer.

        :param name: Printer name - alphanumeric + underscore characters only.
        :param ppd_file: Path to the PPD file with the printer information and configuration.
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        :param info: Human-readable information about the printer.
        :param location: Human-readable printer location info.
        """
        conn = self._get_connection(host, port)
        ppd_file = os.path.abspath(os.path.expanduser(ppd_file))
        conn.addPrinter(name=name, filename=ppd_file, info=info, location=location)

    @action
    def delete_printer(
        self, printer: str, host: Optional[str] = None, port: Optional[int] = None
    ):
        """
        Delete a printer from a CUPS server.

        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        """
        conn = self._get_connection(host, port)
        conn.deletePrinter(printer)

    @action
    def enable_printer(
        self,
        printer: Optional[str],
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Enable a printer on a CUPS server.

        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        """
        conn = self._get_connection(host, port)
        printer = self._get_printer(printer)
        conn.enablePrinter(printer)

    @action
    def disable_printer(
        self,
        printer: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Disable a printer on a CUPS server.

        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        """
        conn = self._get_connection(host, port)
        printer = self._get_printer(printer)
        conn.disablePrinter(printer)

    @action
    def get_jobs(
        self, host: Optional[str] = None, port: Optional[int] = None
    ) -> Dict[int, Dict[str, Any]]:
        """
        Get the list of active jobs.

        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        :return: A job_id -> job_info dict.
        """
        conn = self._get_connection(host, port)
        return conn.getJobs()

    @action
    def accept_jobs(
        self,
        printer: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Start accepting jobs on a printer.

        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        """
        conn = self._get_connection(host, port)
        printer = self._get_printer(printer)
        conn.acceptJobs(printer)

    @action
    def reject_jobs(
        self,
        printer: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Start rejecting jobs on a printer.

        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        """
        conn = self._get_connection(host, port)
        printer = self._get_printer(printer)
        conn.rejectJobs(printer)

    @action
    def cancel_job(
        self,
        job_id: int,
        purge_job: bool = False,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Cancel a printer job.

        :param job_id: Job ID to cancel.
        :param purge_job: Also remove the job from the server (default: False).
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        """
        conn = self._get_connection(host, port)
        conn.cancelJob(job_id, purge_job=purge_job)

    @action
    def move_job(
        self,
        job_id: int,
        source_printer_uri: str,
        target_printer_uri: str,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Move a job to another printer/URI.

        :param job_id: Job ID to cancel.
        :param source_printer_uri: Source printer URI.
        :param target_printer_uri: Target printer URI.
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        """
        conn = self._get_connection(host, port)
        conn.moveJob(
            printer_uri=source_printer_uri,
            job_id=job_id,
            job_printer_uri=target_printer_uri,
        )

    @action
    def finish_document(
        self,
        printer: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Finish sending a document to a printer.

        :param printer: Printer name (default: default configured ``printer``).
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        """
        conn = self._get_connection(host, port)
        printer = self._get_printer(printer)
        conn.finishDocument(printer)

    @action
    def add_printer_to_class(
        self,
        printer_class: str,
        printer: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Add a printer to a class.

        :param printer_class: Class name.
        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        """
        conn = self._get_connection(host, port)
        printer = self._get_printer(printer)
        conn.addPrinterToClass(printer, printer_class)

    @action
    def delete_printer_from_class(
        self,
        printer_class: str,
        printer: Optional[str] = None,
        host: Optional[str] = None,
        port: Optional[int] = None,
    ):
        """
        Delete a printer from a class.

        :param printer_class: Class name.
        :param printer: Printer name.
        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        """
        conn = self._get_connection(host, port)
        printer = self._get_printer(printer)
        conn.deletePrinterFromClass(printer, printer_class)

    @action
    def get_classes(
        self, host: Optional[str] = None, port: Optional[int] = None
    ) -> Dict[str, Dict[str, Any]]:
        """
        Get the list of classes on a CUPS server.

        :param host: CUPS server IP/name (default: default configured ``host``).
        :param port: CUPS server port (default: default configured ``port``).
        :return: dict - class_name -> class_info.
        """
        conn = self._get_connection(host, port)
        return conn.getClasses()


# vim:sw=4:ts=4:et:



from platypush.common.gstreamer import Pipeline
from platypush.plugins.camera import CameraInfo, Camera


class GStreamerCamera(Camera):
    info: CameraInfo
    object: Pipeline


# vim:sw=4:ts=4:et:



import logging

from pychromecast.controllers.media import MediaStatusListener


class SubtitlesAsyncHandler(MediaStatusListener):
    """
    This class is used to enable subtitles when the media is loaded.
    """

    def __init__(self, mc, subtitle_id):
        self.mc = mc
        self.subtitle_id = subtitle_id
        self.initialized = False
        self.logger = logging.getLogger(__name__)

    def new_media_status(self, *_):
        if self.subtitle_id and not self.initialized:
            self.mc.update_status()
            self.mc.enable_subtitle(self.subtitle_id)
            self.initialized = True

    def load_media_failed(self, queue_item_id: int, error_code: int) -> None:
        self.logger.warning(
            "Failed to load media with queue_item_id %d, error code: %d",
            queue_item_id,
            error_code,
        )
